#!/usr/bin/python
""" Define physical constants

All physical constants used in the code belongs to this module. We
also define conventionnal quantities here.
"""

import numpy as np


###############################################
# Definition of constants
###############################################
G = 6.67384e-11  # m^3/kg/s^2
C_LIGHT = 299792458.  # m/s
PARSEC = 3.08567758e16  # m
PROTON_MASS = 1.67262158e-27  # masse du proton kg
H = 6.62617e-34  # J.s
K_BOLTZ = 1.38066e-23  # J/K


#- unit transformation
CLIGHT_A_s  = 2.99792458e18         # [A/s]
CLIGHT_m_s  = 2.99792458e8         # [A/s]
CLIGHT_km_s  = 2.99792458e5         # [A/s]
H_e_s = 6.62606896e-27        # [erg s]

# Stefan-Boltzmann
SIGMA = 2 * np.core.umath.pi ** 5 * K_BOLTZ ** 4 / (15 * H ** 3 * C_LIGHT ** 2)
T_CMB = 2.7255  # Fixen (2009)
RHO_GAMMA = 4 * SIGMA * T_CMB ** 4 / C_LIGHT ** 3  # Energy density of CMB
# today in kg/m3

def hmstodeg(h, m=0, s=0):
    if isinstance(h, str):
        h, m, s = h.split(':')
        h, m, s = float(h), float(m), float(s)
    return h*15 + m*15/60.0 + s*15/3600.0

def dmstodeg(d, m=0, s=0, sign=0):
    if isinstance(d, str):
        d, m, s = d.split(':')
        sign = -1. if d[0] == '-' else 1.
        d, m, s = float(d), float(m), float(s)
    else:
        if not sign:
            if d == 0:
                print('Warning, sign may be ambiguous')
            sign = -1. if d < 0 else 1.
    return (abs(d) + m/60.0 + s/3600.0) * sign

def degtohms(deg):
    hh = deg / 15.0
    h = np.floor(hh)
    m = np.floor((hh - h)*60.0)
    s = (hh - h - m / 60)*3600.0
    return h, m, s

def degtodms(deg):
    si = np.sign(deg)
    hh = abs(deg)
    d = np.floor(hh)
    m = np.floor((hh - d)*60.0)
    s = (hh - d - m / 60)*3600.0
    return si*d, m, s


def flux_ab(wavelength):
    """ Compute AB reference SED at a given wavelength

    Parameters:
    -----------
    wavelength: in Angstrom

    Return:
    -------
    S_ab in erg/(s*cm2*A)
    """
    return 10 ** (-19.44) * C_LIGHT * 1e10 / wavelength ** 2

def hmstodeg(h, m=0, s=0):
    if isinstance(h, str):
        h, m, s = h.split(':')
        h, m, s = float(h), float(m), float(s)
    return h*15 + m*15/60.0 + s*15/3600.0

def dmstodeg(d, m=0, s=0, sign=0):
    if isinstance(d, str):
        d, m, s = d.split(':')
        sign = -1. if d[0] == '-' else 1.
        d, m, s = float(d), float(m), float(s)
    else:
        if not sign:
            if d == 0:
                print('Warning, sign may be ambiguous')
            sign = -1. if d < 0 else 1.
    return (abs(d) + m/60.0 + s/3600.0) * sign

def degtohms(deg):
    hh = deg / 15.0
    h = np.floor(hh)
    m = np.floor((hh - h)*60.0)
    s = (hh - h - m / 60)*3600.0
    return h, m, s

def degtodms(deg):
    si = np.sign(deg)
    hh = np.abs(deg)
    d = np.floor(hh)
    m = np.floor((hh - d)*60.0)
    s = (hh - d - m / 60)*3600.0
    return si*d, m, s



import numpy as np


def reindex(id_array, input=False):
    """
    reindex an indices array to a continuous one

    Parameters
    ----------
    id_array : numpy.array
        Non-continuous index array
    input : bool 
        Wheter the input array is wanted [optional]
    Returns
    -------
    id_array_ri : numpy.array
        Continuous index array
    id_array : numpy.array
        Imput array [if input is True]
    """
    idx = np.arange(len(np.unique(id_array)))
    id_array_unique = np.unique(id_array)
    id_array_ri = np.array([idx[np.where(id_array_unique == i)[0][0]] for i in id_array])
    if input:
        return id_array_ri, id_array
    return id_array_ri



"""
Implementation of 1-D and 2-D B-spline bases.


.. note: on the grids on which the B-Splines are defined
      -----+------+--- ... --+-----+--- ... -----+-----+--
          -k    -k+1         0     1             N    
                             [ specified by user ]
           [       internal, extended grid       ]

.. examples:




"""

import ctypes
import os 
import os.path as op
import numpy as np
from numpy.polynomial.legendre import leggauss
from scipy.sparse import csr_matrix, coo_matrix
import scipy.sparse as sparse
import scipy.sparse.linalg
try:
    from scipy.special import comb
except:
    from scipy.misc import comb

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt



# from . import linearmodels as lm
#from exceptions import LookupError, NotImplementedError

# from IPython import embed 

_lib = np.ctypeslib.load_library('_libnacl', op.dirname(op.abspath(__file__)))
_blossom = _lib.blossom
_blossom.restype = None
_blossom.argtypes = [
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous, writeable'),
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int]

_binary_search_with_guess = _lib.binary_search_with_guess
_binary_search_with_guess.restype = np.int
_binary_search_with_guess.argtypes = [
    ctypes.c_double,
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    ctypes.c_int]

_binary_search = _lib.binary_search
_binary_search.restype = None
_binary_search.argtypes = [
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int]

_blossom_grid = _lib.blossom_grid
_blossom_grid.restype = None
_blossom_grid.argtypes = [
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int]

_deriv_grid = _lib.deriv_grid
_deriv_grid.restype = None
_deriv_grid.argtypes = [
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(np.double, ndim=1, flags='aligned, contiguous'),
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int]


class BSpline(object):
    """1-dimensional B-Spline basis
    
    Implements a 1D-BSpline basis, of arbitrary order, defined on an
    arbitrary grid.
    """
    def __init__(self, grid, order=4):
        self.order=order
        assert self._monotonic_grid(grid)        
        # n_knots is misleading - this should be called n_edges
        # or we could use n_intervals = len(grid) - 1
        self.n_knots = len(grid)
        # if we had n_invervals, then 
        #        self.nj = self.n_intervals + self.order - 1
        # with n_knots, we have:
        self.nj = self.n_knots + self.order - 2
        self.range = grid.min(), grid.max()
        # internally, the grid must be extended
        self._grid = self._extend_grid(grid)
        
    def _monotonic_grid(self, grid):
        d = grid[1:]-grid[:-1]
        if np.sum(d<0) > 0:
            return False
        return True
        
    def _extend_grid(self, grid):
        dx = grid[1]-grid[0]
        p = np.arange(1., self.order)
        pre = -p[::-1] * dx + grid[0]
        dx = grid[-1] - grid[-2]
        post = p * dx + grid[-1]
        return np.hstack((pre, grid, post))

    @property
    def grid(self):
        """the spline intervals, as seen by the user of the class
        """
        return self._grid[self.order-1:-self.order+1]
        
    def __len__(self):
        return self.nj

    def eval(self, x):
        """evaluate the value of the basis functions for each element of x

        Evaluate the 
        
        Args:
            x (ndarray of floats): the values of x

        Returns:
            scipy.sparse.coo_matrix: a sparse, N x p jacobian matrix 
            [N=len(x), p=len(self)] containing the basis values 
            :math:`B_{ij} = B_j(x_i)`

        .. note:  The jacobian matrix triplets are sorted as follows:
            [i=0   j=p      B_p(x_0)      ]  p, lowest integer / B_p(x_i) > 0
            [i=0   j=p+1    B_{p+1}(x_0)  ]
            [i=0   j=p+deg  B_{p+deg}(x_0)]
            [i=1   j=q      B_q(x_1)      ]
                ...    
            The BSpline2.eval function (in fact. BSpline2._cross)
            assumes that order to compute the tensor product of the 1D
            basis elements.
          
            Note also that the old, pure python version of
            CardinalBSpline.eval also returns ordered triplets, but
            the ordering was different.
        """
        nx = len(x)
        N = self.order * nx
        B = np.zeros(N, dtype=np.float64)
        i, j = np.zeros(N, dtype=np.int32), np.zeros(N, dtype=np.int32)
        _blossom_grid(self._grid, len(self._grid), 
                      x, nx, 
                      i, j, B, N, 
                      self.order, self.order, self.nj)
        return coo_matrix((B, (i,j)), shape=(nx, self.nj))
    
    def deriv(self, x, dtype=np.float64):
        """
        evaluate the first derivative of the basis functions for each element of x
        
        Args:
           x: (ndarray of floats) 
       
        Returns:
           a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
           (contains the values of the derivatives)
        """
        nx = len(x)
        N = self.order * nx
        B = np.zeros(N, dtype=np.float64)
        i, j = np.zeros(N, dtype=np.int32), np.zeros(N, dtype=np.int32)
        _blossom_grid(self._grid, len(self._grid), 
                      x, nx, 
                      i, j, B, N, 
                      self.order-1, self.order, self.nj)
        _deriv_grid(self._grid, len(self._grid),
                    i, j, B, N,
                    nx, self.order)        
        return coo_matrix((B, (i,j)), shape=(nx, self.nj))
        

    def deriv_m(self, x, m=2, dtype=np.float64):
        """
        """
        pass
    
    def gram(self, dtype=np.float64):
        """Compute the gramian matrix of the base elements. 

        The gramian is defined by:
        :math:`G_{ij} = \int B_i(x) B_j(x) dx`

        Args:
           dtype: Gramian dtype (np.float32 or np.float64)

        Returns:
           scipy.sparse.coo_matrix: a sparse p x p matrix [p=len(self)]
                                   containing the gramian values
        """
        p, w = leggauss(self.order)
        g = self._grid
        # linear transform grid-elements -> [-1,1]
        ak = np.repeat(0.5 * (g[1:]-g[:-1]), self.order)
        bk = np.repeat(0.5 * (g[1:]+g[:-1]), self.order)
        # integration grid 
        nk = len(g) - 1
        pp = np.tile(p, nk)
        pp = ak*pp+bk
        # weights
        ww = np.tile(w, nk)
        N = len(ww)
        i = np.arange(N)
        W = coo_matrix((ak*ww, (i,i)), shape=(N,N))
        B = self.eval(pp)
        return B.T * W * B

    def __call__(self, x, p, deriv=0):
        if deriv == 0:
            B = self.eval(x)
        elif deriv == 1:
            B = self.deriv(x)
        else:
            B = self.deriv_m(x, m=deriv)
        return B * p

    def linear_fit(self, x, y, w=None):
        # we now use cholmod, since it behaves much better than
        # scipy.sr parse.linalg.spsolve when the size of the basis increases
        # this is especially important in 2D (see below)
        J = self.eval(x) # .tocsr()
        if w is not None:
            J.data *= w[J.row]
            yy = w * y
        else:
            yy = y
        factor = cholesky_AAt(J.T)
        return factor(J.T * yy)


class BSpline2D(object):
    """
    2D BSpline basis of arbitrary order, 
    defined on an arbitrary grid

    The basis is the cross-product of two 1D BSpline bases, along x and y:

    .. math::
       B_{ij} = B_i(x) \\times B_j(y)
    """
    def __init__(self, gx, gy, x_order=4, y_order=4):
        self.bx = BSpline(gx, order=x_order)
        self.by = BSpline(gy, order=y_order)
        self.nj = self.bx.nj * self.by.nj
        
    def __len__(self):
        """
        return the size of the basis
        """
        return self.nj

    def _cross(self, N, ix, jx, vx, iy, jy, vy):
        """
        compute the cross-product: 

        .. math::
           B_{ij}(x,y) = B_i(x) \times B_j(y)

        Args:
          N: (int) 
             number of points (N=len(x))
          i: ndarray of ints 
             a ndarray containing the row-indices in the jacobian matrix
          jx: ndarray of ints 
              the column indices in the matrix returned by the x-basis 
          vx: ndarray of floats, 
              the values B_j(x)
          jy: ndarray of ints 
              the column indices in the matrix returned by the y-basis 
          vy: ndarray of floats
              the values B_j(y)

        Returns:
          J: (scipy.sparse.coo_matrix) 
              the values of the cross-product as a (N,n) sparse jacobian matrix, 
              [N is the number of points, n the size of the 2D-basis]

        Note:
          This implementation makes an assumption on how the return value 
          of ``BSpline.eval'' is sorted internally.  
          See the documentation of this routine above.           

        TODO: 
          One could gain ~ a factor 2 in execution time, by implementing 
          the tile's and repeat's in C.
        """
        
        xo, yo = self.bx.order, self.by.order
        
        i_   = ix.reshape(-1,xo).repeat(yo,axis=1).ravel()
        jx_  = jx.reshape(-1,xo).repeat(yo,axis=1).ravel()
        vx_  = vx.reshape(-1,xo).repeat(yo,axis=1).ravel()
        jy_  = np.tile(jy.reshape(-1,yo), (1,xo)).ravel()
        vy_  = np.tile(vy.reshape(-1,yo), (1,xo)).ravel()        
        data_ = vx_ * vy_
        j_ = jy_ * self.bx.nj + jx_
        return coo_matrix((data_, (i_,j_)),
                          shape=(N,self.nj))
    
    def eval(self, x, y):
        """
        evaluate and return the values of the basis functions for (x,y)
        
        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          B: (scipy.sparse.coo_matrix)
              a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
              containing the basis values: B_{ij} = B_j(x_i)
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        Jx = self.bx.eval(x)
        ix, jx, vx = Jx.row, Jx.col, Jx.data
        Jx = None
        Jy = self.by.eval(y)        
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None        
        #        embed()
        r = self._cross(N, ix, jx, vx, iy, jy, vy)
        return r
        
    def gradient(self, x, y):
        """
        evaluate and return the derivatives vs. x and y of the basis functions for (x,y)
        
        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          dvdx: (scipy.sparse.coo_matrix)
                 a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                 containing the values: B_{ij} = B_j'(x_i) * B_j(y_i)
          dvdy: (scipy.sparse.coo_matrix)
                 a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                 containing the values: B_{ij} = B_j(x_i) * B_j'(y_i)
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        
        Jxp = self.bx.deriv(x)
        ix, jx, vx = Jxp.row, Jxp.col, Jxp.data
        Jxp = None
        Jy  = self.by.eval(y)
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        ddx = self._cross(N, ix, jx, vx, iy, jy, vy)
        
        ix = jx = vx = None
        iy = jy = vy = None
        
        Jx  = self.bx.eval(x)
        ix, jx, vx = Jx.row, Jx.col, Jx.data
        Jx = None
        Jyp = self.by.deriv(y)
        iy, jy, vy = Jyp.row, Jyp.col, Jyp.data
        Jyp = None
        ddy = self._cross(N, ix, jx, vx, iy, jy, vy)
        
        return ddx, ddy        
        
    def hessian(self, x, y):
        raise NotImplementedError('BSpline2D.hessian not implemented (yet)')

    def laplacian(self, x, y):
        raise NotImplementedError('BSpline2D.hessian not implemented (yet)')

    def __call__(self, x, y, p, deriv=0):
        if deriv == 0:
            B = self.eval(x,y)
        elif deriv == 1:
            B = self.gradient(x,y)
        else:
            raise exceptions.NotImplementedError('m>1 derivatives not implemented')
        return B * p

    def linear_fit(self, x, y, v, w=None):
        """fit the basis coefficients on the data passed in argument

        Perform a least square fit of the basic coefficients. 

        Args:
            x (ndarray of floats): x coordinates
            y (ndarray of floats): y coordinates
            v (ndarray of floats): values

        Returns:
            ndarray of floats: fit solution, i.e. basis coefficients

        .. note: ~ 12 seconds on a Intel(R) Core(TM) i5-2540M CPU @ 2.60GHz
                 for 10^6 values and a 10x10 basis. Dominated by the cholesky
                 factorization.
        """
        J = self.eval(x,y)
        if w is not None:
            J.data *= w[J.row]
            vv = w * v
        else:
            vv = v
        factor = cholesky_AAt(J.T)
        return factor(J.T * vv)

class CardinalBSplineC(object):
    """
    New CardinalBSpline basis, with a blossoming 
    function implemented in C. 
    
    (Simpler, and faster in most cases)
    """
    def __init__(self, n=10, x_range=(0., 1.), order = 4):
        self.order = order 
        self.n_knots = n
        self.nj = self.n_knots + self.order - 1
        self.range = x_range
        self.dx = (x_range[1]-x_range[0]) / n 
        
    def __len__(self):
        return self.nj 

    @property
    def grid(self):
        r = np.arange(self.range[0], self.range[1]+0.5*self.dx, self.dx)
        return r
    
    def _int_grid(self, x):
        """
        return a copy of the x' array, 
        scaled so that the knots are at integer locations 
        """
        return (x-self.range[0]) / self.dx + self.order - 1.
    
    def eval(self, x):
        """
        evaluate the value of the basis functions for each element of x
        
        Args:
          x : (ndarray of floats) 
              the x-values 
       
        Returns:
          B: (scipy.sparse.coo_matrix) 
                a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                containing the basis values: B_{ij} = B_j(x_i)        
                
        TODO: 
          add a float32 version (?)
        """
        nx = len(x)
        N = self.order * nx
        B = np.zeros(N, dtype=np.float64)
        xi = self._int_grid(x)
        i, j = np.zeros(N, dtype=np.int32), np.zeros(N, dtype=np.int32)
        _blossom(xi, nx, i, j, B, N, self.order, self.order, self.nj)
        return coo_matrix((B, (i,j)), shape=(nx, self.nj))
            
    def deriv(self, x):
        """
        evaluate the first derivative of the basis functions for each element of x
        
        Args:
           x: (ndarray of floats) 
       
         Returns:
            a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
            (contains the values of the derivatives)
        """
        nx = len(x)
        N = self.order * nx
        B = np.zeros(N)
        xi = self._int_grid(x)
        i, j = np.zeros(N, dtype=np.int32), np.zeros(N, dtype=np.int32)
        _blossom(xi, nx, i, j, B, N, self.order-1, self.order, self.nj)
        B[:-1] -= B[1:]
        return coo_matrix((B, (i,j)), shape=(nx, self.nj)) / self.dx

    def deriv_m(self, x, m=2, dtype=np.float64):
        """
        evaluate the m-th first derivative of the basis functions for each element of x
        
        Args:
           x: (ndarray of floats) 
       
         Returns:
            a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
            (contains the values of the derivatives)
        """
        nx = len(x)
        N = self.order * nx
        B = np.zeros(N, dtype=dtype)
        xi = self._int_grid(x)
        i, j = np.zeros(N, dtype=np.int32), np.zeros(N, dtype=np.int32)
        if (self.order - m) <= 0:
            return coo_matrix((B, (i,j)), shape=(nx,self.nj)) / self.dx
        _blossom(xi, nx, i, j, B, N, self.order-m, self.order, self.nj)
        d = np.arange(1,m+1) # was m+1
        Cmp = comb(m, d)
        Cmp[::2] *= -1.
        BB = B.copy()
        for d,c in zip(d,Cmp):
            BB[:-d] += c * B[d:]
        return coo_matrix((BB, (i,j)), shape=(nx,self.nj)) / self.dx

    def gram(self):
        """
        Compute the gramian matrix of the base elements:
        
        .. math::
             G_{ij} = \int B_i(x) B_j(x) dx

        """
        nk = self.n_knots
        N = self.order * (self.n_knots+1)
        p, w = leggauss(self.order)
        # I am here ... 
        grid = np.linspace(self.range[0], self.range[1], self.n_knots+1)
        _integ()
        W = coo_matrix((W, (i,i)), shape=(N,N))
        B = self.eval(pp)
        return B.T * W * B        
    
    # def __call__(self, x, p, deriv=0):
    #     if deriv == 0:
    #         B = self.eval(x)
    #     elif deriv == 1:
    #         B = self.deriv(x)
    #     else:
    #         B = self.deriv_m(x, m=deriv)
    #     return B * p
    
    def linear_fit(self, x, y):
        J = self.eval(x)
        JtJ = J.T.dot(J)
        return scipy.sparse.linalg.spsolve(JtJ, J.T.dot(y))


class CardinalBSpline(object):
    """
    Implements a 1D-BSpline basis of arbitrary order, 
    defined on a grid of evenly spaced knots.
    """
    def __init__(self, n=10, x_range=(0.,1.), order=4):
        """
        By default, compute cardinal BSplines of order 4 
        (i.e. CubicBSplines)
        """
        self.order = order 
        self.n_knots = n
        self.nj = self.n_knots + self.order - 1
        self.range = x_range
        self.dx = (x_range[1]-x_range[0]) / n 

    def __len__(self):
        """
        return the size of the basis
        """
        return self.nj

    @property
    def grid(self):
        r = np.arange(self.range[0], self.range[1]+0.5*self.dx, self.dx)
        return r                 
    
    def _start(self, xi, dtype):
        """
        compute the 1-st order splines given the xi (step functions)
        
        Args:
          xi: (ndarray)
              array of x's, rescaled so that the knots are at integer positions. 

        Returns:
          a tuple (k, xi, i, j, b, om) where 
            - k is the order of the spline (1) 
            - xi are the rescaled x's (reference to the arguments)
            - i  are the line indexes of the jacobian matrix (i = arange(len(x)))
            - j  are the column indexes of the jacobian matrix 
            - b  are the spline values 
            - om are the recursion coefficients: \omega_{j,k+1} = (x - j) / k
        """
        N = len(xi)
        i = np.arange(N)
        j = np.floor(xi)
        b = np.ones(N)
        om = (xi - j).astype(dtype)
        return [(1, xi, i, j, b, om)]    
    
    def _merge(self, xxx_todo_changeme):
        """
        auxiliary function, called while blossoming up. 
        
        compute the B^k spline from a pair of adjacent B^{k-1}
        splines. The recursion formula is:
        
        .. math::
             B^k_{j}(x) = \omega^k_j(x) B^{k-1}_{j}(x) + (1-\omega^{k}_{j+1}) B^{k-1}_{j+1}(x)

        with 

        .. math::
             \omega^k_j(x) = \frac{x-t_j}{t_{j+k-1}-t_{j}}
             
        Args:
            left (tuple or None):  (k, xi, i,m j, b, om)
            right (tuple or None): (k, xi, i,m j, b, om)
              
        Returns:
            new tuple, of the form (k+1, xi, i, j, b, om) of order k+1
        """
        (left, right) = xxx_todo_changeme
        if right is None:
            k, xi, i, j, b, om = left
            b = b * om
            om = (xi - j) / (k+1.)
            return (k+1, xi, i, j, b, om)
        elif left is None:
            k, xi, i, j, b, om = right
            b = b * (1.-om)
            om = (xi - j + 1.) / (k+1.)
            return (k+1, xi, i, j-1, b, om)
        k_l, xi, i, j_l, b_l, om_l = left
        k_r, xi, i, j_r, b_r, om_r = right
        b = om_l*b_l + (1.-om_r)*b_r
        om = (xi-j_l) / (k_l + 1.)
        return (k_l+1, xi, i, j_l, b, om)
    
    def _blossom(self, bvlist):
        """
        blossom: compute recursively the k-th order splines from the (k-1)-th order splines.
        """
        l = [None] + bvlist + [None]
        l = list(zip(l[:-1], l[1:]))
        return [self._merge(b) for b in l]
    
    def get_xi(self, x):
        """
        return a copy of the x' array, 
        scaled so that the knots are at integer locations 

        The spline corresponding to the first node is at location i=0.
        This means that x.min() is at location i=(order-1).
        """
        return (x-self.range[0]) / self.dx + self.order - 1.
    
    def eval(self, x, dtype=np.float32):
        """
        evaluate the value of the basis functions for each element of x
        
        Args:
          x : (ndarray of floats) 
              the x-values 
       
        Returns:
          B: (scipy.sparse.coo_matrix) 
                a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                containing the basis values: B_{ij} = B_j(x_i)
                
        Note:
          This method is used by the 2D-version of the CardinalBSplines, which 
          assumes that the triplets (i,j,v) are sorted like they are here, i.e.: 
 
                [i=0, j=0, B_0(x0)]
                [i=1, j=0, B_0(x1)]
                      ...
                [i=0, j=1, B_1(x0)]
                      ...          
        """
        xi = (x-self.range[0]) / self.dx + self.order - 1.
        s = self._start(xi, dtype)
        for i in range(1,self.order):
            s = self._blossom(s)
        i = np.hstack([b[2] for b in s])
        j = np.hstack([b[3] for b in s])
        v = np.hstack([b[-2] for b in s])
        J = coo_matrix((v, (i,j)), dtype=dtype, shape=(len(x), self.nj))
        return J
    
    def deriv(self, x, dtype=np.float32):
        """
        evaluate the derivatives of the basis functions for each element of x
        
        Args:
           x: (ndarray of floats) 
       
         Returns:
            a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
            (contains the values of the derivatives)        
        """
        xi = (x-self.range[0]) / self.dx + self.order - 1.

        # lower order splines (k-1)
        s = self._start(xi, dtype)
        for i in range(1,self.order-1):
            s = self._blossom(s)
            
        # add 
        N = len(x) ; z = np.zeros(N)
        k0, xi, i, j0, b0, om0 = s[0]
        k1, xi, i, j1, b1, om1 = s[-1]
        s = [(k0, xi, i, j0-1, z, z)] + s + [(k1, xi, i, j1+1, z, z)]
            
        # and compute the differences B_j^k-1 - B_j+1^k-1
        l = list(zip(s[0:-1], s[1:]))
        s = []
        for low, high in l:
            kl, xi, i, jl, bl, oml = low
            kh, xi, i, jh, bh, omh = high
            s.append((kl, xi, i, jl, bl-bh, oml))
            
        i = np.hstack([b[2] for b in s])
        j = np.hstack([b[3] for b in s])
        v = np.hstack([b[-2] for b in s])
        J = coo_matrix((v, (i,j)), dtype=dtype, shape=(N, self.nj))
        # checked with numerical derivative that the self.dx is really needed
        return J / self.dx  

    def deriv_m(self, x, m=2, dtype=np.float32):
        """
        evaluate the m-th derivatives of the basis functions for each element of x
        
         Args:
          x: (ndarray of floats) 
       
         Returns:
           a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
            (contains the values of the derivatives)        
        """
        xi = (x-self.range[0]) / self.dx + self.order - 1.

        # splines of order k-m 
        s = self._start(xi, dtype)
        for i in range(1,self.order-m):
            s = self._blossom(s)
            
        # add m zero elements before and after the structure 
        N = len(x) ; z = np.zeros(N)
        k0, xi, i, j0, b0, om0 = s[0]
        k1, xi, i, j1, b1, om1 = s[-1]
        s_pre = []
        s_post = []
        for q in range(m):
            s_pre.append([k0, xi, i, j0-m+q, z, z])
            s_post.append([k0, xi, i, j1+q+1, z, z])
        s = s_pre + s + s_post
        #        s = [(k0, xi, i, j0-1, z, z)]*m + s + [(k1, xi, i, j1+1, z, z)]*m # BUG !

        # now, we want to compute  \sum_p=0^m f[p] B_{j+p,k-m}(x)
        # with f[p] = (-1)^p C_m^p 
        # First, we compute the bvfilter f[p] itself 
        f = comb(m, np.arange(m+1))
        f[1::2] *= -1. 
        # then, we stack the basis values into one single array 
        l = []
        bvalues = np.vstack([u[-2] for u in s]).T
        for k in range(len(s)-m):
            kl, xi, i, jl, bl, oml = s[k]
            bv = np.sum(f * bvalues[:,k:k+m+1], axis=1)
            l.append((kl, xi, i, jl, bv, oml))
            
        i = np.hstack([b[2] for b in l])
        j = np.hstack([b[3] for b in l])
        v = np.hstack([b[-2] for b in l])
        J = coo_matrix((v, (i,j)), dtype=dtype, shape=(N, self.nj))
        return J / self.dx

    def gram(self, dtype=np.float64):
        """
        Compute the gramian matrix of the base elements:

        .. math::
            G_{ij} = \int B_i(x) B_j(x) dx
        """
        nk = self.n_knots        
        # Gaussian quadrature points and weights
        p, w = leggauss(self.order) 
                
        # points
        #        x = np.arange(self.range[0], self.range[1]+self.dx, self.dx)
        x = np.linspace(self.range[0], self.range[1], self.n_knots+1)
        ak = np.repeat(0.5*(x[1:]-x[:-1]), self.order)
        bk = np.repeat(0.5*(x[1:]+x[:-1]), self.order)
        pp = np.tile(p, nk)
        pp = ak*pp+bk
        
        # weights
        ww = np.tile(w, nk)
        N = len(ww)
        i = np.arange(N)
        W = coo_matrix((ak*ww, (i,i)), shape=(N,N))
        
        B = self.eval(pp, dtype=dtype)
        return B.T * W * B
                
    def integral(self, x):
        """
        Also have a look at this: 
        http://www.sciencedirect.com/science/article/pii/S009630030500189X
        http://imamat.oxfordjournals.org/content/17/1/37.abstract
        """
        pass

    def __call__(self, x, b, deriv=0, dtype=np.float32):
        """
        Syntaxic sugar so that if bs is a bspline
        bs(x, b) returns the evaluated spline function on the grid x
        with spline parameters b
        
        deriv: set it to the degree of the derivation you expect.
        deriv=0 is similar default and returns the spline       
        """
        if deriv == 0:
            A = self.eval(x, dtype=dtype)
        elif deriv == 1:
            A = self.deriv(x, dtype=dtype)
        else:
            A = self.deriv_m(x, m=deriv, dtype=dtype )

        return (A * b)

    def linear_fit(self, x, y):
        """
        Makes a linear fit of the spline 
        
        x is the list of abscissa (not necessarily ordered)
        y is the corresponding list of data

        Make sure that all x values are within the bounds of the bspline you defined.        

        FIXME: add the weights
        """
        # essential to have float64, otherwise the solver complains
        J = self.eval(x, dtype=np.float64)
        JtJ = J.T.dot(J)
        return sparse.linalg.spsolve(JtJ, J.T.dot(y))
        #        return np.linalg.solve((A.T * A).todense(), A.T * y)
        

def gram(basis1, basis2, dtype=np.float64):
    """
    Compute the gramian matrix of elements of the two (different) bases
    ``basis1`` and ``basis2``. 
    
    .. math::  G_{ij} = \int B_i(x) C_j(x) dx

    where B_i and C_j are elements of the two bases. 
    """
    # Gaussian quadrature points and weights
    deg = basis1.order + basis2.order
    p, w = leggauss(deg)
    
    # points -- use the basis grids ! 
    x1 = np.linspace(basis1.range[0], basis1.range[1], basis1.n_knots+1)
    x2 = np.linspace(basis2.range[0], basis2.range[1], basis2.n_knots+1)
    x = np.hstack((x1,x2)) ; x.sort() ; x = np.unique(x)
    nk = len(x)-1
    ak = np.repeat(0.5*(x[1:]-x[:-1]), deg)
    bk = np.repeat(0.5*(x[1:]+x[:-1]), deg)
    pp = np.tile(p, nk)
    pp = ak*pp+bk
    
    # weights
    ww = np.tile(w, nk)
    N = len(ww)
    i = np.arange(N)
    W = coo_matrix((ak*ww, (i,i)), shape=(N,N))
    
    # some bases accept dtypes, other not.. 
    # TODO: uniformize this (ban np.float32 and remove dtype options ?)
    B1 = basis1.eval(pp)
    B2 = basis2.eval(pp)
    return B1.T * W * B2


def lgram(spectrum_basis, filter_basis, **keys):
    """Compute the :math:`\lambda-`Gramian of both bases.

    When computing broadband fluxes, using transmission functions as
    spectra developed on spline bases, one needs to compute the
    following quantities: 

    .. math:: 
         G_{ij} = \int B_i(\lambda) B_j(\lambda) \lambda d\lambda
      
    This function computes the G-matrix above, and returns it as a
    sparse matrix. The spectrum basis may be shifted by a factor
    (1+z), for example:

    .. math::
         G_{ij} = \int B_i(\lambda) B_j(\lambda / (1+z)) \lambda d\lambda

    """
    # Gaussian quadrature points and weights
    deg = spectrum_basis.order + filter_basis.order
    p, w = leggauss(deg)
    
    z = keys.get('z', 0.)
    to_photons = keys.get('to_photons', False)
    lambda_power = keys.get('lambda_power', 1)
    
    # points -- use the basis grids ! 
    if z > 0.:
        x1 = spectrum_basis.grid * (1.+z)
    else:
        x1 = spectrum_basis.grid / (1.-z)
    x2 = filter_basis.grid
    x = np.hstack((x1,x2)) ; x.sort() ; x = np.unique(x)
    nk = len(x)-1
    ak = np.repeat(0.5*(x[1:]-x[:-1]), deg)
    bk = np.repeat(0.5*(x[1:]+x[:-1]), deg)
    pp = np.tile(p, nk)
    pp = ak*pp+bk
    
    # weights
    ww = np.tile(w, nk)
    N = len(ww)
    i = np.arange(N)
    
    A_hc = 50341170.081942275 / (1.+z) ** 2 
    W = coo_matrix((ak * ww * A_hc * np.power(pp, lambda_power), (i,i)), 
                   shape=(N,N))
    
    if z > 0.:
        B1 = spectrum_basis.eval(pp / (1.+z))
    else:
        B1 = spectrum_basis.eval(pp / (1.-z))
    B2 = filter_basis.eval(pp)
    return B1.T * W * B2


_leggauss_cache = {}

def integ(basis, n=0):
    """
    Compute the integral (of n-th moments) of the spline basis functions.
    
    .. math::
       I_i^{[n]} = \int x^n B_i(x) dx
    
    Args:
      basis (bspline basis): 1-D bspline basis to integrate 
      n (int): moment order 

    Returns:
      a vector of length `len(basis)` containing the n-th moment of each function.      
    """
    global _leggauss_cache
    g = basis.grid
    aa = np.matrix(0.5 * (g[1:]-g[:-1]))
    bb = np.matrix(0.5 * (g[1:]+g[:-1]))
    nk = len(g) - 1

    deg = basis.order + n
    if deg in _leggauss_cache:
        x,w = _leggauss_cache[deg]
    else:
        x,w = leggauss(basis.order + n)
        _leggauss_cache[deg] = x,w
        
    xx = np.array(np.dot(aa.T, np.matrix(x)) + bb.T).flatten().squeeze()
    dx = np.array(np.dot(aa.T, np.matrix(np.ones(len(x))))).flatten().squeeze()
    
    J = basis.eval(xx)
    w = np.tile(w, nk) * xx**n * dx
    i,j = np.zeros(len(w)), np.arange(len(w))
    W = coo_matrix((w,(i,j)), shape=(1,len(w)))

    I = (W * J).tocoo()

    return I.data[I.col]



class CardinalBSpline2D(object):
    """
    2D-BSpline basis of arbitrary order, 
    defined on a grid of evenly spaced knots.
    
    The basis is the cross-product of two BSpline bases, along x and y:
    
    .. math::
      B_{ij}(x,y) = B_i(x) \times B_j(y)
      
    """
    def __init__(self, 
                 nx=10, x_range=(0.,1.), x_order=4,
                 ny=10, y_range=(0.,1.), y_order=4):
        self.bx = CardinalBSpline(n=nx, x_range=x_range, order=x_order)
        self.by = CardinalBSpline(n=ny, x_range=y_range, order=y_order) # there is a bug here, no ? 
        self.nj = self.bx.nj * self.by.nj

    def __len__(self):
        """
        return the size of the basis
        """
        return self.nj

    def _cross(self, N, i, jx, vx, jy, vy, dtype=np.float32):
        """
        compute the cross-product: B_{ij}(x,y) = B_i(x) \times B_j(y)

        Args:
          N: (int) 
             number of points (N=len(x))
          j: ndarray of ints 
             a ndarray containing the row-indices in the jacobian matrix
          jx: ndarray of ints 
              the column indices in the matrix returned by the x-basis 
          vx: ndarray of floats, 
              the values B_j(x)
          jy: ndarray of ints 
              the column indices in the matrix returned by the y-basis 
          vy: ndarray of floats
              the values B_j(y)

        Returns:
          J: (scipy.sparse.coo_matrix) 
              the values of the cross-product as a (N,n) sparse jacobian matrix, 
              [N is the number of points, n the size of the 2D-basis]

        Note:
          This implementation makes an assumption on how the return value 
          of ``CardinalBSpline.eval'' is sorted internally.  
          See the documentation of this routine above.           
        """
        i  = i.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        jx = jx.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        vx = vx.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        jy = np.tile(jy.reshape(-1,N).T, (1,self.by.order))
        vy = np.tile(vy.reshape(-1,N).T, (1,self.by.order))
        
        data = vx * vy
        i = i.ravel()
        j = jy.ravel()*self.bx.nj + jx.ravel()
        
        return coo_matrix((data.ravel(), (i,j)), 
                          shape=(N,self.nj), dtype=dtype)        

    def eval(self, x, y, dtype=np.float32):
        """
        evaluate and return the values of the basis functions for (x,y)
        
        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          B: (scipy.sparse.coo_matrix)
              a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
              containing the basis values: B_{ij} = B_j(x_i)
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        Jx = self.bx.eval(x, dtype=dtype)
        ix, jx, vx = Jx.row, Jx.col, Jx.data
        Jx = None
        Jy = self.by.eval(y, dtype=dtype)        
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        #        embed()
        return self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)


    def gradient(self, x, y, dtype=np.float32):
        """
        evaluate and return the derivatives vs. x and y of the basis functions for (x,y)
        
        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          dvdx: (scipy.sparse.coo_matrix)
                 a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                 containing the values: B_{ij} = B_j'(x_i) * B_j(y_i)
          dvdy: (scipy.sparse.coo_matrix)
                 a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
                 containing the values: B_{ij} = B_j(x_i) * B_j'(y_i)
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        
        Jxp = self.bx.deriv(x, dtype=dtype)
        ix, jx, vx = Jxp.row, Jxp.col, Jxp.data
        Jxp = None
        Jy  = self.by.eval(y, dtype=dtype)
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        ddx = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        ix = jx = vx = None
        iy = jy = vy = None
        
        Jx  = self.bx.eval(x, dtype=dtype)
        ix, jx, vx = Jx.row, Jx.col, Jx.data
        Jx = None
        Jyp = self.by.deriv(y, dtype=dtype)
        iy, jy, vy = Jyp.row, Jyp.col, Jyp.data
        Jyp = None
        ddy = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        return ddx, ddy        

    def hessian(self, x, y, dtype=np.float32):
        """
        evaluate and return the partial second derivatives of the
        basis functions.

        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          d2Bdx2, d2Bdy2, d2Bdxy : (scipy.sparse.coo_matrix)
              a tuple of 3 sparse, N x p jacobian matrices [N=len(x), p=len(self)]
              containing second derivatives of the basis values.
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        
        Jxp = self.bx.deriv_m(x, m=2, dtype=dtype)
        ix, jx, vx = Jxp.row, Jxp.col, Jxp.data
        Jxp = None
        Jy  = self.by.eval(y, dtype=dtype)
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        ddx2 = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        ix = jx = vx = None
        iy = jy = vy = None
        
        Jx = self.bx.eval(x, dtype=dtype)
        ix, jx, vx = Jx.row, Jx.col, Jx.data 
        Jx = None
        Jyp = self.by.deriv_m(y, m=2, dtype=dtype)
        iy, jy, vy = Jyp.row, Jyp.col, Jyp.data
        Jyp = None
        ddy2 = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)

        ix = jx = vx = None
        iy = jy = vy = None
        
        Jxp = self.bx.deriv_m(x, m=1, dtype=dtype)
        ix, jx, vx = Jxp.row, Jxp.col, Jxp.data
        Jxp = None
        Jyp = self.by.deriv_m(y, m=1, dtype=dtype)
        iy, jy, vy = Jyp.row, Jyp.col, Jyp.data
        Jyp = None
        ddxy = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        return ddx2, ddy2, ddxy
                

    def laplacian(self, x, y, dtype=np.float32):
        """
        evaluate and return the laplacian of the basis functions. 

        Args:
          x: (ndarray of floats) 
              x-coordinates of the entry points
          y: (ndarray of floats) 
              y-coordinates of the entry points

        Returns:
          L : (scipy.sparse.coo_matrix)
              a sparse, N x p jacobian matrix [N=len(x), p=len(self)]
              containing the laplacian of basis values: L_{ij} = \Delta B_j(x_i)
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        
        Jxp = self.bx.deriv_m(x, m=2, dtype=dtype)
        ix, jx, vx = Jxp.row, Jxp.col, Jxp.data
        Jxp = None
        Jy  = self.by.eval(y, dtype=dtype)
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        ddx2 = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        ix = jx = vx = None
        iy = jy = vy = None
        
        Jx = self.bx.eval(x, dtype=dtype)
        ix, jx, vx = Jx.row, Jx.col, Jx.data 
        Jx = None
        Jyp = self.by.deriv_m(y, m=2, dtype=dtype)
        iy, jy, vy = Jyp.row, Jyp.col, Jyp.data
        Jyp = None
        ddy2 = self._cross(N, ix, jx, vx, jy, vy, dtype=dtype)
        
        return ddx2 + ddy2    
    
    def eval_old(self, x, y):
        """
        Still here, but deprecated. Will disappear soon. 
        """
        if len(x) != len(y): 
            raise ValueError('x and y should have the same length')        
        N = len(x)
        Jx = self.bx.eval(x)
        ix, jx, vx = Jx.row, Jx.col, Jx.data
        Jx = None
        ix = ix.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        jx = jx.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        vx = vx.reshape(-1,N).T.repeat(self.bx.order,axis=1)
        
        Jy = self.by.eval(y)        
        iy, jy, vy = Jy.row, Jy.col, Jy.data
        Jy = None
        #        iy = np.tile(iy.reshape(-1,N).T, (1,4))
        jy = np.tile(jy.reshape(-1,N).T, (1,self.by.order))
        vy = np.tile(vy.reshape(-1,N).T, (1,self.by.order))
        
        data = vx * vy
        i = ix.ravel()
        j = jy.ravel()*self.bx.nj + jx.ravel()

        return coo_matrix((data.ravel(), (i,j)), 
                          shape=(N,self.nj), dtype=np.float32)


class CubicBSpline:
    """Define a basis of (h+4)) uniform cubic B-splines on a
    segment [0,Lx]
    """
    def __init__(self, N=1, Lx=1.0, Ly=1.0):
        self.Lx = Lx
        self.N = N + 3
        self.dx = float(Lx) / N
        self.tx = np.arange(-3 * self.dx, Lx + 4 * self.dx, self.dx)
        self.compute_weights()

    def compute_weights(self):
        self.w = np.ones(5)
        self.ti = np.arange(5)
        for i in range(5):
            for j in range(5):
                if j != i:
                    self.w[i] = self.w[i] * (self.ti[j] - self.ti[i])
        self.w = 1.0 / self.w

    def b3(self, t):
        b = np.zeros(t.shape)
        for i in range(5):
            cond = t >= self.ti[i]
            _t = t[cond] - self.ti[i]
            _t *= _t * _t
            b[cond] += self.w[i] * _t
        return b

    def b3_eval(self, t):
        i = np.vstack([np.floor(t).astype('int') + e for e in range(0, 4)])
        val = np.vstack([self.b3(t - i[e, :] + 3) for e in range(4)])
        return i, val

    def _eval(self, x):
        i, M = self.b3_eval(x / self.dx)
        return i, M

    def sparse_mat(self, x, factor=None):
        Ncell = len(x)
        i, val = self._eval(x)
        index_x = np.tile(np.arange(Ncell), 4)
        index_y = i.flatten()
        if factor is not None:
            val *= factor
        index = index_y < self.N
        return index_x[index], index_y[index], val.flatten()[index]

    def eval(self, pars, x):
        from scipy import sparse as sp
        index_x, index_y, val = self.sparse_mat(x.flatten())
        A = sp.coo_matrix((val, (index_x, index_y)),
                          #shape=(index_x.max() + 1, len(self.tx) - 1)).tocsr()
                          ).tocsr()
        return x, np.reshape(A * pars, x.shape)

    # def fit(self, x, y, w=None, full_output=False):
    #     M = lm.LinearSparseModel()
    #     M.models = {"pars": lambda: self.sparse_mat(x)}
    #     if w is not None:
    #         M.w = w
    #     M.B = y
    #     if full_output:
    #         return M.solve(), M
    #     else:
    #         return M.solve()

class BicubicBSpline(CubicBSpline):
    """Define a basis of (h+4)(k+4) uniform bicubic B-splines on a
    rectangle [0,Lx]x[0,Ly].
    """
    def __init__(self, M=1, N=1, Lx=1.0, Ly=1.0):
        self.Lx = Lx
        self.Ly = Ly
        self.dx, self.dy = float(Lx) / M, float(Ly) / N
        self.tx = np.arange(-3 * self.dx, Lx + 4 * self.dx, self.dx)
        self.ty = np.arange(-3 * self.dy, Ly + 4 * self.dy, self.dy)
        self.compute_weights()

    def eval(self, x, y):
        if len(x) != len(y):
            raise ValueError('x and y should have same length')
        i, M = self.b3_eval(x / self.dx)
        j, N = self.b3_eval(y / self.dy)
        _m, _n = np.mgrid[0:4, 0:4]
        _n = _n.flatten()
        _m = _m.flatten()
        val = M[_m, :] * N[_n, :]
        return i[_m, :], j[_n, :], val

    def sparse_mat(self, x, y, factor=None):
        Ncell = len(x)
        i, j, val = self.eval(x, y)
        if factor is not None:
            val *= factor
        index_x = np.tile(np.arange(Ncell), 16)
        index_y = i.flatten() + j.flatten() * (len(self.tx) - 4)
        return index_x, index_y, val.flatten()

    def show(self, pars, res=[100j, 100j], xy=None):
        from scipy import sparse as sp
        #import pylab
        if xy is None:
            _x = np.linspace(0, self.Lx - 1, abs(res[0]) + 2)[1:-1]
            _y = np.linspace(0, self.Ly - 1, abs(res[1]) + 2)[1:-1]
        #x,y=np.mgrid[0:self.Lx-1:res[0],0:self.Ly-1:res[1]]
            x, y = np.meshgrid(_x, _y)
        else:
            x, y = xy
        index_x, index_y, val = self.sparse_mat(x.flatten(), y.flatten())
        A = sp.coo_matrix((val, (index_x, index_y))).tocsr()
        return x, y, np.reshape(A * pars, x.shape)
        #pylab.imshow(np.reshape(A*pars,(100,100)))
        #pylab.colorbar()

    # def b3_integral(self, a, b):
    #      return 0.25*(self.w*((b-self.ti)**4*(b>=self.ti) - (a-self.ti)**4*(a>=self.ti))).sum()

    # def integral(self, i, j, quad):
    #     a,b,c,d= quad
    #     i1 = self.dx*self.b3_integral((a-self.tx[i])/self.dx, (b-self.tx[i])/self.dx)
    #     i2 = self.dy*self.b3_integral((c-self.ty[j])/self.dy, (d-self.ty[j])/self.dy)
    #     return i1*i2



    # def plot_debug(self, x, y):
    #     import pylab as pl 
    #     pl.figure()        
    #     Jx = self.bx.eval(x)
    #     p = np.random.choice([0.,1.], size=len(self.bx))
    #     pl.plot(x, Jx * p, 'r.')
        
    #     pl.figure()
    #     Jy = self.by.eval(y)
    #     p = np.random.choice([0.,1.], size=len(self.by))
    #     pl.plot(y, Jy * p, 'r.')        

    #     pl.figure()
    #     ax = pl.subplot(221)
    #     pl.plot(x, 'r.')
    #     ay = pl.subplot(222)
    #     pl.plot(y, 'b.')
    #     pl.subplot(223, sharex=ax)
    #     pl.plot(Jx.col, 'r.')
    #     pl.subplot(224, sharex=ay)
    #     pl.plot(Jy.col, 'b.')
        
    #     pl.figure()
        
    #     ix = Jx.row.reshape(-1,N).T.repeat(self.by.order,axis=1)
    #     jx = np.repeat(Jx.col, self.by.order)
    #     vx = np.repeat(Jx.data, self.by.order)

    #     iy = np.tile(Jy.row, self.bx.order)
    #     jy = np.tile(Jy.col, self.bx.order)
    #     vy = np.tile(Jy.data, self.bx.order)

    #     j = jx * self.by.nj + jy
    #     #        j = jy * self.bx.nj + jx
    #     N,n = len(x), self.nj
    #     J = coo_matrix((vx*vy, (ix,j)), shape=(N,n))
        
    #     pl.subplot(421)
    #     pl.plot(x, 'r.')
    #     pl.subplot(422)
    #     pl.plot(y, 'b.')
    #     pl.subplot(423)
    #     pl.plot(ix, 'r.')
    #     pl.subplot(424)
    #     pl.plot(iy, 'b.')
    #     pl.subplot(425)  
    #     pl.plot(jx, 'r.')      
    #     pl.subplot(426)  
    #     pl.plot(jy, 'b.')      
    #     pl.subplot(427)  
    #     pl.plot(vx, 'r.')      
    #     pl.subplot(428)  
    #     pl.plot(vy, 'b.')      
    #     return J



import numpy as np
from ..dataset import TrainingDataset
from ..util import salt2

from math import *
import astropy.cosmology as cosmology


class Generator(object):
    r"""
    Data generator.
    Generate ideal simulation.

    * Parameters of SNe
    For a fixed number of supernovae, we uniformly draw redshifts between :math:`z = 0.001` and :math:`z = 0.8`
    by default.
    The training batches are truncated in redshift with respect to the cosmological batches
    in order to have a complete training batch, without selection bias.

    Once the redshifts are obtained, we compute the luminosity distance defined as :

    .. math::
        d_L(z) =  c\int^{z}_{0} \frac{d\tilde{z}}{H(\tilde{z})} (1+z)

    where :

    .. math::
        H^{2}(z) = H_0^2 \left[ \Omega_{m} (1+z)^3 + \Omega_{r} (1+z)^4 +  \Omega_{k} (1+z)^2 +
        \Omega_{DE} (1+z)^{3(1+w)} \right]

    Our fiduciary cosmology is the following Flat$-\Lambda$CDM cosmology~:
    with :

    .. math::
         \bullet \, & H_0 = 70, \mathrm{km.s^{-1}.Mpc^{-1}} \\
         \bullet \, & \Omega_m = 0.25 \\
         \bullet \, & w = -1 \\
         \bullet \, & \Omega_DE = 0.75 \\
         \bullet \, & \Omega_k = 0. \\
         \bullet \, & \Omega_\gamma = 5*10^{-5} \\

    Once the distances are known, we can simulate the eigen-parameters of the SNe.
    We first generate the values of :math:`X_1` and :math:`c` according to the two following normal laws~:

    .. math::
        X_1 \sim \mathcal{N}(0, 1.) \quad \quad \quad c \sim \mathcal{N}(0, 0.1)

    then we calculate :math:`X_0` according to the following formula:

    .. math::
        X_0 = \frac{10^{-4}}{d_L(z)^2} 10^{-0.4 (\tilde{M}_{B^\star} - \alpha X_1 + \beta c)}

    The factor :math:`10^{-4}` corresponds to the passage of the square of the luminosity
    distance from :math:`kpc^2 \,  to \, pc^2`.
    For the standardization parameters, we take those determined by the JLA~ analysis:

    .. math::
         \bullet \, & M_{B^\star} = -19.0906 \\
         \bullet \, & \alpha = 0.13 \\
         \bullet \, & \beta = 3 \\

    If LSST survey is simulated filters are : :math:`u, g,r,i,z`
    If not, observing band depend on the z.
    If :math:`z` < 0.05, bands are u, B, V and r from the SWOPE instrument.
    If 0.05 < :math:`z` < 0.2, they are griz from SDSS's instrument.
    and if 0.2 < :math:`z`, bands are griz of Megacam.


    Attributes
    ----------
    n_sn : int
        Number of SNe generated
    n_sp_sn : int
        Number of Spectra per sn, common to all sne.
    n_lc_sn : int
        Number of light curve POINT per sne, common to all light curves.
    n_band : int
        Number of different filter use (only not LSST)
    lc : numpy.rec.array
        Light curves data with data_type as type.
    sp : numpy.rec.array
        Spectral data, with data_type as type.
    z : numpy.array
        Redshift of each data point (photometric then spectroscopic)
    cosmo : astropy.cosmology.FlatLambdaCDM
        Cosmology
    Mb : float
        SN Ia absolute magnitude
    alpha : float
        Parameter of brighter-slower relation
    beta : float
        Parameter of brighter-bluer relation
    flux_at_10pc : float
        Flux of SN observed at 10 pc.
    X0_norm : float
       Model overall X0 normalisation.
    snInfo : numpy.rec.array
        Array of SNe information (:math:`z`, :math:`tmax`, :math:`x1`, :math:`x0`, :math:`c`)
    n_filter : int
        Number of different filter use (only not LSST)
    bands : list
        Bands used in the survey.
    remove_band : list
        if band should be removed.
    lsst_simu : bool
        if sample generated using LSST filter.
    seed : int
        Seed for cadence.
    z_range : list
        Redshift range for the training sample
    data_type : list
        NaCl data type.
    snInfo_type : list
        Type of information of SNe.
    sigma_lc : float
        Dispersion of photometric data (only not LSST)
    sigma_sp : float
        Dispersion of spectral data.
    trainingDataset: nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    """

    def __init__(self, n_sn, n_sp_sn=1, n_lc_sn=20, n_band=4,
                 z_range=[0.01, 0.85], remove_band=[],
                 lsst_simu=True, seed=0, sigma_lc=0.05, sigma_sp=0.02,
                 date_evenly_sparse=False):
        """
        Constructor - computes the arguments

        Parameters
        ----------
        n_sn : int
            Number of SNe generated
        n_sp_sn : int
            Number of Spectra per sn, common to all sne.
        n_lc_sn : int
            Number of light curve POINT per sne, common to all light curves.
        n_band : int
            Number of different filter use (only not LSST)
        z_range : list
            redshift range for the training sample
        remove_band : list
            if band should be removed.
        lsst_simu : bool
            if sample generated using LSST filter.
        seed : int
            Seed for cadence.
        sigma_lc : float
            Dispersion of photometric data (only not LSST)
        sigma_sp : float
            Dispersion of spectral data.
        date_evenly_sparse :
            If observation date are uniformly distributed.
        """
        self.nb_sn = n_sn
        self.n_filter = n_band
        
        self.n_sp_sn = n_sp_sn
        self.n_lc_sn = n_lc_sn

        self.lc, self.sp, self.z = None, None, None
        self.cosmo = None
        self.Mb, self.alpha, self.beta = None, None, None
        self.flux_at_10pc, self.X0_norm = None, None
        self.snInfo, self.n_filter, self.bands = None, None, None

        # for LSST
        self.lsst_simu = lsst_simu
        self.remove_band = remove_band
        self.bands_survey()
        
        self.seed = seed
        np.random.seed(self.seed)
        
        self.z_range = z_range
        self.data_type = [('Date', '<f8'), ('Flux', '<f8'), ('FluxErr', '<f8'),
                          ('Filter', '|S20'), ('Wavelength', '<f8'), ('MagSys', '|S20'),
                          ('ZHelio', '<f8'), ('sn_id', '<i4'), ('spec_id', '<i4')]
        self.snInfo_type = [('z', '<f8'), ('tmax', '<f8'), ('x1', '<f8'),
                            ('x0', '<f8'), ('c', '<f8')]

        self.init_info()
        # print("0", self.snInfo['x1'].sum())
        self.init_photometric_data()
        # print("1", self.snInfo['x1'].sum())
        self.init_spectral_data(date_evenly_sparse)
        # print("2", self.snInfo['x1'].sum())
                
        self.sigma_sp = sigma_sp
        self.sigma_lc = sigma_lc
        self.indexing()
        self.trainingDataset = TrainingDataset(self.lc, self.sp, self.snInfo)

    def indexing(self):
        """
        Reindex spectral ('spec_id', 'sn_id') and photometric data ('band_id', 'lc_id', 'sn_id'),
        when data have been removed.
        """
        n_lc, n_sp = len(self.lc), len(self.sp)
        # band indexation
        dict_bd = {}
        _, idx_bd = np.unique(self.lc['Filter'], return_index=True)
        for i_bd, bd in enumerate(self.lc['Filter'][np.sort(idx_bd)]):
            dict_bd[bd] = i_bd
        id_bd = np.array([dict_bd[bd] for bd in self.lc['Filter']])

        # light curve indexation
        c = 0
        id_lc = np.ones(len(self.lc['Flux']))
        
        for i in range(self.lc['sn_id'][-1]+1):
            idx_sn = self.lc['sn_id'] == i
            lcs = self.lc[idx_sn]
            _, idx = np.unique(lcs["Filter"], return_index=True)
            for bd_sn in lcs['Filter'][np.sort(idx)]:
                id_lc[(self.lc['sn_id'] == i) & (self.lc['Filter'] == bd_sn)] = c
                c += 1
        
        id_lc = np.hstack(np.array(id_lc))
        self.lc = np.lib.recfunctions.rec_append_fields(self.lc, names=['lc_id', 'band_id', 'i'],
                                                        data=[id_lc.astype(int), id_bd, np.arange(n_lc)])
        i_sp = n_lc + np.arange(n_sp)
        sp_ones = np.ones_like(i_sp)
        self.sp = np.lib.recfunctions.rec_append_fields(self.sp, names=['lc_id', 'band_id', 'i'],
                                                        data=[-1*sp_ones, -1 * sp_ones, n_lc+np.arange(n_sp)])

    def init_info(self, tmax_dist=[15., 30.], x1_dist=[0, 1], c_dist=[0, .1]):
        r"""
        SN parameters initialisation:
        :math:`X_1` : follow a centred and reduced normal distribution,
        :math:`c` : follow a centred normal distribution of std = 0.1,
        :math:`t_{max}` : follow in uniform distribution between 15 and 30 MJD,
        :math:`z` : follow also a uniform distribution.

        :math:`X_0`, is defined, using the other parameters, as follow :
                
        .. math:
            X_0 = 10^{-0.4(M_b-\alpha x1 + \Beta c )} \frac{(10pc)^2}{d_L^2(z)}

        dL is calculated following a flat :math:`\Lambda CDM` and
        :math:`M_b`, :math:`\alpha`, :math:`\beta` are taken from JLA.
        """
        n_sn = self.nb_sn
        np.random.seed(self.seed)
        z = np.random.uniform(self.z_range[0], self.z_range[1], size=floor(n_sn))

        color = np.random.normal(c_dist[0], c_dist[1], size=n_sn)
        sig = np.random.normal(x1_dist[0], x1_dist[1], size=n_sn)
        sig /= sig.std()
        print('X1 removing mean and correct from standard deviation; \n c removing mean')
        sig -= sig.mean()
        color -= color.mean()
 
        np.random.seed(self.seed)
        tmax = np.random.uniform(tmax_dist[0], tmax_dist[1], size=n_sn)
        self.cosmo = cosmology.FlatLambdaCDM(H0=70, Om0=0.25)
        dl = np.array(self.cosmo.luminosity_distance(z)) * 1.E3 # from Mpc to kpc
        self.Mb = -19.0906
        self.alpha = 0.13
        self.beta = 3.0

        self.flux_at_10pc = np.power(10., -0.4 * self.Mb)
        self.X0_norm = self.flux_at_10pc * 1.E-4  # from kpc^2 to (10pc)^2
        norm = self.X0_norm / dl**2
        norm *= np.power(10., 0.4*(self.alpha*sig - self.beta*color))

        self.snInfo = np.rec.fromarrays((z, tmax, sig, norm.ravel(), color),
                                        names=['z', 'tmax', 'x1', 'x0', 'c'])
        self.z = z
        print(f"color m : {np.mean(self.snInfo['c'])}, s : {np.std(self.snInfo['c'])}")
        print(f"x1 m : {np.mean(self.snInfo['x1'])}, s : {np.std(self.snInfo['x1'])}")

    def init_spectral_data(self, date_evenly_sparse,
                           range_spectra=[2000, 11000]):
        r"""
        Observation initialisation, each spectra range is drawn 
        from an int between 2000 and 5000, 
        finish between 6000 and 9000 math:`\AA`.
        Observation date follow a centred normal distribution of std = 10 days.

        The resolution is constant and equal to 10 math:`\AA`.
        Each spectrum is transform in observer frame, by multiplying
        the wavelength by a factor math:`1+z` and adding to the observation date, math:`t_{max}` and 
        times math:`1+z`

        Parameters
        ----------
        date_evenly_sparse : bool
             If observation spectral date should be uniformly distributed.
             Else follow gumbel distribution.
        range_spectra : list,
             range of generated spectra
        """
        n_sn = self.nb_sn
        np.random.seed(self.seed)
        
        nsp = np.random.randint(self.n_sp_sn, self.n_sp_sn + 1, size=n_sn)
        sn_sp = np.array([[i]*nsp[i] for i in range(n_sn)]).ravel()

        date_sp, wave_sp, zhelio_sp, obs_id_sp, sn_id_sp = [],  [], [], [], []
        dates = np.linspace(-20, 45, nsp.sum())  # 50
        np.random.shuffle(dates)
            
        for i in range(len(sn_sp)):
            r0 = np.random.randint(range_spectra[0], 3000)  # 5000
            r1 = np.random.randint(8000, range_spectra[1])  # 6000
            n_pts_wl = ceil((r1-r0)/10) 

            wave_sp.append(np.linspace(r0, r1, n_pts_wl))
            # if date_evenly_sparse:
            #     date_sp.append([dates[i]] * n_pts_wl)
            # else:
            ddate = np.random.gumbel(-2, 8., 1)
            if (ddate > 35) ^ (ddate < -15):  # 35
                # print(ddate[0])
                # print('\n removing spectra bigger than 35 smaller than -15 !!! \n')
                # np.random.seed(self.seed)
                ddate = np.random.normal(0, 5, 1)
                # print(ddate[0])                    
            date_sp.append([ddate[0]] * n_pts_wl)  # np.random.normal(3, 15, 1)[0]] * n_pts_wl)
                
            obs_id_sp.append([i] * n_pts_wl)
            sn_id_sp.append([sn_sp[i]] * n_pts_wl)
            
        sn_id_sp = np.hstack(sn_id_sp)
        zhelio_sp = self.z[sn_id_sp]
        # date_sp = (np.hstack(date_sp) + self.snInfo['tmax'][sn_id_sp]) * (1 + zhelio_sp)
        date_sp = (np.hstack(date_sp)) * (1 + zhelio_sp) + self.snInfo['tmax'][sn_id_sp]
        wave_sp = np.hstack(wave_sp) * (1 + zhelio_sp)
        obs_id_sp = np.hstack(obs_id_sp)
        
        sp_none = np.array([b''] * len(date_sp))       
        flux_sp = np.zeros_like(date_sp)
        
        self.sp = np.rec.fromarrays((date_sp, flux_sp, flux_sp,  sp_none, wave_sp,
                                     sp_none, zhelio_sp, sn_id_sp, obs_id_sp),
                                    dtype=self.data_type)

    def init_photometric_data(self):
        r"""
        Initiate photometric cadence. For each SN n_band band of observation, depending on the redshift
        or common and equal to LSST bands.

        Date of observation are evenly sample between -15 and 30 days, with a gaussian noise. 
        """
        n_sn = self.nb_sn
        self.n_filter = self.n_filter
        
        nlc = np.random.randint(self.n_lc_sn, self.n_lc_sn + 1, size=n_sn)
        sn_lc = np.array([[i] * nlc[i] * self.n_filter for i in range(len(nlc))]).ravel()
        zhelio_lc = self.z[sn_lc]
        # date_lc = (np.array([[ii for ii  in np.linspace(-15,30,nlc[i]) + np.random.normal(0, 0.5, 1)]*
        # self.n_filter for i in range(len(nlc))]).ravel() + self.snInfo['tmax'][sn_lc]) * (1 + zhelio_lc)
        date_lc = (np.array([[ii for ii in np.linspace(-15, 35, nlc[i]) + np.random.normal(0, 0.5, 1)] *
                             self.n_filter for i in range(len(nlc))]).ravel()) * \
                  (1 + zhelio_lc) + self.snInfo['tmax'][sn_lc]

        band_lc = self.make_band(nlc)
        # self.make_band(nlc)
        # band_lc = self.bands_lc_obs
        flux_lc = np.zeros_like(date_lc)
        lc_none = np.array([b''] * len(date_lc))
        lc_ones = np.ones_like(sn_lc)
        lc_mean_wavelength = np.zeros_like(flux_lc)
        bands = [i.decode('UTF-8') for i in np.unique(band_lc)]
        filterset = salt2.load_filters(bands)
        for bd in bands:
            idxmw = band_lc == bd.encode('UTF-8')
            lc_mean_wavelength[idxmw] = filterset.mean_wavelength([bd])
        
        self.lc = np.rec.fromarrays((date_lc, flux_lc, flux_lc,  band_lc,
                                     lc_mean_wavelength, lc_none, zhelio_lc, sn_lc,
                                     -1 * lc_ones),  # obs_id_lc),
                                    dtype=self.data_type)

    def bands_survey(self):
        r"""
        Attribute band to the light curves, if LSST all SNe are observed through griz.
        If not depending on z, they are observed through CSP, SDSS and SNLS bands.
        """
        if self.lsst_simu:
            self.bands = ['LSST::u', 'LSST::g', 'LSST::r', 'LSST::i', 'LSST::z']
            if self.remove_band != 0:
                for bd in self.remove_band:
                    self.bands.remove(bd)
            self.n_filter = len(self.bands)
        else:
            self.bands = ['SWOPE::u', 'SWOPE::V', 'SWOPE::B', 'SWOPE::r',
                          'SDSS::g', 'SDSS::r', 'SDSS::i', 'SDSS::z',
                          'MEGACAMPSF::g', 'MEGACAMPSF::i', 'MEGACAMPSF::r', 'MEGACAMPSF::z']

    def make_band(self, nlc):
        r"""
        Create the band array of each photometric observation.
        If observations are made through LSST cadence, all sne are observe by math:`u,g,r,i,z` bands.
        If not, observing band depend on the z.
        If z < 0.05, bands are u, B, V and r from the SWOPE instrument.
        If 0.05 < z < 0.2, they are griz from SDSS's instrument.
        and if 0.2 < z, bands are griz of Megacam.

        Parameters
        ----------
        nlc : array
            Number of observation per Lcs.

        Returns
        -------
        band_lc : array
            Band corresponding to each observation
        """
        band_lc = []
        if self.lsst_simu:
            for iz in range(len(self.z)):
                band_lc0 = np.array(np.repeat(self.bands, nlc[iz]), dtype='|S16')
                band_lc.append(band_lc0)
            band_lc = np.hstack(band_lc)
        else:
            for iz in range(len(self.z)):
                zz = self.z[iz]
                if zz < 0.05:
                    i_band = 0
                elif (zz >= 0.05) & (zz < 0.2):
                    i_band = 1
                else:
                    i_band = 2
                band_lc0 = np.array(np.repeat(self.bands[i_band * self.n_filter:(i_band + 1) * self.n_filter],
                                              nlc[iz]), dtype='|S16')
                band_lc.append(band_lc0)
            band_lc = np.hstack(band_lc)
        return band_lc



#!/usr/bin/python

""" Provides:
- Func1D: easy handling of tabulated 1D functions
"""

import numpy as np
from scipy import interpolate, integrate
import matplotlib.pyplot as plt
import sys


class Func1D(object):
    """ Wrapper around different ways of representing 1D function

    Func1D provides:
    - a callable interface to tabulated data (through 1D interpolation,
    - multiplication and addition of Func1D objects,
    - A routine for integration in a range.
    """
    def __init__(self, *func_like, **kwargs):
        """ Provide a Func1D interface over different data types

        Parameters:
        -----------
        func_like can be either:
            - a scalar
            - a callable (with only 1 parameter)
            - two arrays: in that case the function is an interpoltor of y = f(x) 
                          where x is the first array and y the second array


        It has an x_max and x_min member, that can either be set by options, 
        or that are the bounds of the x array. By default, they are both 0. 
        All this is made to be similar to snfit Sample1dFunction

        When 2 Func1D instances are multiplied x_min = max of the 2 x_min
        and x_max = min of the 2 x_max, like in sampledfunction.cc
        """
        self.x_min = 0
        self.x_max = 0

        if "x_min" in kwargs:
            self.x_min = kwargs["x_min"]
        if "x_max" in kwargs:
            self.x_max = kwargs["x_max"]

        if not "bounds_error"  in kwargs:
            bounds_error=False
        else:
            bounds_error = kwargs["bounds_error"]

        if len(func_like) > 2:
            raise ValueError("WARNING: Func1D can only deal with 1 or 2 args, not more")

        elif len(func_like) == 2:
            self.func = interpolate.interpolate.interp1d(
                func_like[0], func_like[1],
                bounds_error=bounds_error,
                fill_value=0)
            self.x_min = np.min(func_like[0])
            self.x_max = np.max(func_like[0])
        else:
            func_like = func_like[0]

            if isinstance(func_like, Func1D):
                self.func = func_like.func
            elif np.core.numeric.isscalar(func_like):
                self.func = lambda x: func_like
            elif callable(func_like):
                self.func = func_like
            else:
                raise ValueError("Don't know how to create a Func1D with what you provided, exiting")

    def __call__(self, x):
        """ 1D function evaluation at x.

        Parameter:
        ----------
        x: scalar or 1D array
        """
        return self.func(x)

    def __mul__(self, func2):
        """ Return a Func1D that will evaluate to the product of self
        and f2.
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: self.func(x) * Func1D( func2 ).func(x), x_min=x_min, x_max=x_max )

    #- Func1D multiplication is commutative
    __rmul__ = __mul__

    def __div__(self, func2):
        """ Return a Func1D that will evaluate to the product of self
        and f2.
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: self.func(x) / Func1D( func2 ).func(x), x_min=x_min, x_max=x_max )
    
    def __add__(self, func2):
        """ Return a Func1D that will evaluate to the sum of self
        and func2.
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: self.func(x) + Func1D( func2 ).func(x), x_min=x_min, x_max=x_max )

    #- Func1D addition is commutative
    __radd__ = __add__


    def __sub__(self, func2):
        """ Return a Func1D that will evaluate to the difference of self
        and func2.
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: self.func(x) - Func1D( func2 ).func(x), x_min=x_min, x_max=x_max )

    def __rsub__(self, func2):
        """ Return a Func1D that will evaluate to the difference of func2 and self

        func2 - self
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: Func1D( func2 ).func(x) - self.func(x), x_min=x_min, x_max=x_max )

    
    def __pow__(self, func2):
        """ Return a Func1D that will evaluate to the self ** func2
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x:  self.func(x) ** (Func1D( func2 ).func(x)), x_min=x_min, x_max=x_max )

    def __rpow__(self, func2):
        """ Return a Func1D that will evaluate to the func2 ** self
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        return Func1D( lambda x: (Func1D( func2 ).func(x)) ** self.func(x), x_min=x_min, x_max=x_max )
    
    
    
    
    def integrate(self, range_inf, range_sup, method=100):
        """ Evaluate function integral in a range.

        Parameters:
        -----------
        range_inf: float
                   lower limit of the integration range
        range_sup: float
                   upper limit of the integration range
        method: can be either
              "romb": The integral is evaluated using the romberg method
              N (int): The integral is evaluated using a N points
                       simpson quadrature.
        """
        if isinstance(method, int):
            x = np.core.function_base.linspace(range_inf, range_sup, method)
#            return integrate.quadrature.simps(self(x), x=x)
            return integrate.simps(self(x), x=x)
        elif method == 'romb':
#            return integrate.quadrature.romberg(self, range_inf, range_sup)
            return integrate.romberg(self, range_inf, range_sup)

    def mean(self, range_inf=None, range_sup=None, method=100, order=0):
        """ Evaluate the mean using the function as a weight.

        Parameters:
        -----------
        range_inf: float
                   lower limit of the integration range
        range_sup: float
                   upper limit of the integration range
        method: to evaluate the integrals. Can be either:
              "romb": The integral is evaluated using the romberg method
              N (int): The integral is evaluated using a N points
                       simpson quadrature.
        """
        def identity(x):
            return x

        if range_inf is None:
            range_inf = self.x_min
        if range_sup is None:
            range_sup = self.x_max
        
        m = (self * identity).integrate(range_inf, range_sup, method=method)
        d = self.integrate(range_inf, range_sup, method=method)
        m /= d
        if order == 0:
            return m
        elif order == 1:
            def f(x):
                return np.abs(x - m)
        else:
            def f(x):
                return (x - m) ** order
        m2 = (self * f).integrate(range_inf, range_sup, method=method)
        return m2 / d

    def _x(self, x=None):
        if x is None:
            x = np.linspace(self.x_min, self.x_max, 100)
        return x
    
    def max(self, x=None):
        return self(self._x(x)).max()

    def min(self, x=None):
        return self(self._x(x)).min()
    
    def plot(self, start=None, stop=None, num=50, norm=False, scale=1.,  **keys):
        """ Plot the function in a range.

        Parameters:
        -----------
        start: The lower bound of the range.
        stop: The upper bound of the range.
        num: number of evaluation points in the range.
        norm: to normalize by the max value
        scale: to rescale the plot function

        All other keywords parameters are passed directly to plot.

        Example:
        --------
        >>> f = Func1D(np.sin)
        >>> f.plot(-np.pi, np.pi, 100, color='k')
        """
        if start is None:
            start = self.x_min
        if stop is None:
            stop = self.x_max
        x = np.core.function_base.linspace(start, stop, num)
        if norm:
            y = self(x) * scale
            return plt.plot(x, y / y.max(), **keys)
        else:
            return plt.plot(x, self(x) * scale, **keys)


if __name__ == "__main__":
    #- Implements few tests that I feel better if I can run from time to time

    x = Func1D(np.arange(1., 10., 0.1), np.arange(1., 10., 0.1))
    y = Func1D(np.arange(1., 10., 0.1), np.arange(1., 10., 0.1)+10.)
    

    print(("x(4.) + 1.    = %f     (should be 5.)" % ((x + 1.)(4.))))
    print(("1.    + x(4.) = %f     (should be 5.)" % ((1. + x)(4.))))

    print(("x(4.) - 1.    = %f     (should be 3.)" % ((x - 1.)(4.))))
    print(("1.    - x(4.) = %f     (shoudld be -3.)" % ((1. - x)(4.))))


    print(("x(4.) * 2.    = %f     (should be 8,)" % ((x * 2.)(4.))))
    print(("2.    * x(4.) = %f     (should be 8.)" % ((2. * x)(4.))))

    print(("x(4.) ** 2.    = %f     (should be 16,)" % ((x ** 2.)(4.))))
    print(("2.    ** x(4.) = %f     (should be 16.)" % ((2. ** x)(4.))))

    print(("(x + y)(2.)    = %f     (should be 14.)" % ((x + y)(2.))))
    print(("(y + x)(2.)    = %f     (should be 14.)" % ((y + x)(2.))))
    print(("(x * y)(2.)    = %f     (should be 24.)" % ((x * y)(2.))))
    print(("(y * x)(2.)    = %f     (should be 24.)" % ((y * x)(2.))))
    print(("(x ** y)(2.)   = %f     (should be 4096.)" % ((x ** y)(2.))))
    print(("(y ** x)(2.)   = %f     (should be 144.)" % ((y ** x)(2.))))    
    

    



"""

TODO: Put the iterative training steps.
"""
import numpy as np
import os 
from .lib import bspline
from .minimizers import Minimizer, ModelPulls
from .models.constraints import Constraints2D
from .models.regularizations import SplineRegul2D
from .models.variancemodels import SimpleVarianceModel

from .plotting import plots

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt


class FitModel2D(object):
    """
    Fitting procedure class.
    From a simulation, a model, error models, and parameters,
    can fit the model (given a minimizer) on data.
    If parameters are fixed, regularization and constraints are adjusted.

    All the parameters of the fit, amplitude of constraints, regularization can be specified.

    Attributes
    ----------
    training_dataset : nacl.dataset.TrainingDataset
        Training dataset.
    model : nacl.salt
        Model.
    gamma_init : numpy.array ou float
        Parameters used to initialize the error snake.
    basis_knot : 2-list
        Number of wavelength and photometric knot.
    init_from_salt2_file : str
        If the model is to be initialized with the original
        SALT2-4 surfaces and bases, get them from this file.
    norm_band : str
        Use this band to normalize the model.
    degree : int
        Degree of the spectrum recalibration polynomials
    BETA : float
        Diagonal offset in hessian matrix where use of cholesky.
    queue_wsp : list
        Model queue of eval unit.
    pars_free : list
        List of free parameters in the fit.
    beta : numpy.array
        Fitted model parameter.
    beta0 : numpy.array
        Model parameter initialization.
    reg : nacl.models.regularizations
        Class of regularization.
    pars_generation : numpy.array
        Parameters used for the simulation.
    all_pars : list
        List of all the parameters of the model.
    cons : nacl.models.constraints
        Class of constraints.
    x1_var_real : float
        Amplitude of constraints to the total :math:`\chi^2`
    N_sn_4plot : int
        Index of SN data to plot.
    variance_model : nacl.models.variancemodels.SimpleVarianceModel
        Error snake model.
    gamma : numpy.array
        Fitted error snake parameter :math:`\gamma`
    gamma0 : numpy.array
        Error snake parameter :math:`\gamma` at initialisation
    kappa_init : numpy.array
        Color scatter parameters.
    color_scatter_func : object
        Color scatter function used.
    color_scatter : nacl.models.variancemodels.ColorScatter
        Color scatter model.
    mu_pq : float
        Amplitude of the linear constraints to the total :math:`\chi^2`.
    log : list
        For debugging reason, stored at each step all fit quantities.
    N_ITER : int
        Maximum number of fit iteration.
    df_stop : float
        :math:`\Delta \chi^2` after which the fit stops.
    pars_fix : list
        List a parameter fixed during the fit.
    mu_pq_var_x1 : float
        Amplitude of constraints on :math:`X_1` variance to the total :math:`\chi^2`
    mu_reg : float
        Amplitude of the regularization to the total :math:`\chi^2`
    chi2_normalization : float
        Ratio of :math:`\chi^2` of spectral data and :math:`\chi^2` of photometric data.
    minimize_func : nacl.minimizers.Minimizer
        Minimizer.
    log : list
        For debugging reason, stored at each step all fit quantities.
    df_stop : float
        :math:`\Delta \chi^2` after which the fit stops.
    pars_fix : list
        List a parameter fixed during the fit.

    """
    def __init__(self, training_dataset, model_func, normalization_band_name='SWOPE::B',
                 init_from_salt2_file='data/salt2.npz', pars_fix=[],
                 minimizer=Minimizer, degree=3, disconnect_sp=False,
                 gamma_init=0.05, kappa_init=None, x1_var_real=False,
                 mu_pq=1e10, mu_pq_var_x1=None, mu_reg=1,
                 variance_model_bins=(3,3),
                 chi2_normalization=1, color_scatter_func=None, beta=1e-10,
                 log=True, df_stop=1e-1, n_iter=500, basis_knot=[127, 20],
                 pars_generation=None):
        """
        Constructor.

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
            Training dataset.
        model_func : nacl.salt
            Model.
        gamma_init : numpy.array ou float
            Parameters used to initialize the error snake.
        basis_knot : 2-list
            Number of wavelength and photometric knot.
        init_from_salt2_file : str
            If the model is to be initialized with the original
            SALT2-4 surfaces and bases, get them from this file.
        normalization_band_name : str
            Use this band to normalize the model.
        degree : int
            Degree of the spectrum recalibration polynomials
        beta : float
            Diagonal offset in hessian matrix where use of cholesky.
        pars_generation : numpy.array
            Parameters used for the simulation.
        x1_var_real : float
            Amplitude of constraints to the total :math:`\chi^2`
        N_sn_4plot : int
            Index of SN data to plot.
        kappa_init : numpy.array
            Color scatter parameters.
        color_scatter_func : nacl.models.variancemodels.ColorScatter
        mu_pq : float
            Amplitude of the linear constraints to the total :math:`\chi^2`.
        log : list
            For debugging reason, stored at each step all fit quantities.
        n_iter : int
            Maximum number of fit iteration.
        df_stop : float
            :math:`\Delta \chi^2` after which the fit stops.
        pars_fix : list
            List a parameter fixed during the fit.
        mu_pq_var_x1 : float
            Amplitude of constraints on :math:`X_1` variance to the total :math:`\chi^2`
        mu_reg : float
            Amplitude of the regularization to the total :math:`\chi^2`
        variance_model_bins : tuple(int, int)
            Number of bins of the vraince surface
        chi2_normalization : float
            Ratio of :math:`\chi^2` of spectral data and :math:`\chi^2` of photometric data.
        minimizer : nacl.minimizers.Minimizer
            Minimizer.
        log : list
            For debugging reason, stored at each step all fit quantities.
        df_stop : float
            :math:`\Delta \chi^2` after which the fit stops.
        pars_fix : list
            List a parameter fixed during the fit.

        """
        self.training_dataset = training_dataset
        self.norm_band = normalization_band_name
        self.degree = degree
        self.BETA = beta
        self.gamma_init = gamma_init
        self.kappa_init = kappa_init
        self.color_scatter_func = color_scatter_func
        self.x1_var_real = x1_var_real

        self.pars_free, self.x0, self.g0, self.reg, self.cons = None, None, None, None, None

        self.basis_knot = basis_knot
        self.model = model_func(self.training_dataset,
                                normalization_band_name=self.norm_band,
                                spectrum_recal_degree=self.degree,
                                basis_knot=self.basis_knot)
      
        self.queue_wsp = self.model.queue.copy()

        self.init_model_pars(filename=init_from_salt2_file, sn_info=self.training_dataset.sne)

        self.variance_model = SimpleVarianceModel(self.model, disconnect_sp=disconnect_sp,
                                                  gamma_init=self.gamma_init,
                                                  bins=variance_model_bins)

        self.pars_generation = pars_generation
        self.N_sn_4plot = None
        self.all_pars = ['X0', 'tmax', 'X1', 'c', 'M0', 'M1', 'CL',
                         'SpectrumRecalibration', 'eta_calib', 'kappa_color']
        
        self.mu_pq = mu_pq
        self.mu_pq_var_x1 = mu_pq_var_x1
        self.mu_reg = mu_reg
        self.chi2_normalization = chi2_normalization
        self.minimize_func = minimizer

        self.log = log
        self.N_ITER = n_iter
        self.df_stop = df_stop
        
        self.pars_fix = pars_fix
        self.pars_release()
        self.model.pars0 = self.model.pars.full[:].copy()
        
        
    def init_model_pars(self, filename, sn_info):
        r"""
        Initiate model parameters.

        Parameters
        ----------
        filename : str
            The original
            SALT2-4 surfaces and bases, get them from this file.
        sn_info : numpy.rec.array
            SNe information (:math:`(z, X_0, X_1, c, t_{max})`)
        """
        m0, m1, cl = self.init_with_salt2(filename, self.model)
        try:
            self.model.pars['X0'][:] = sn_info['x0']
            self.model.pars['X1'][:] = sn_info['x1']
        except KeyError:
            self.model.pars['X0'][:] = sn_info['norm']
            self.model.pars['X1'][:] = sn_info['sig']
        self.model.pars['tmax'][:] = sn_info['tmax']
        self.model.pars['c'][:] = sn_info['c']
        
        self.model.pars['M0'][:] = m0
        self.model.pars['M1'][:] = m1
        self.model.pars['CL'][:] = cl
        self.model.norm = self.model.normalization(band_name=self.norm_band)

    @staticmethod
    def init_with_salt2(filename, model):
        """
        Initiate the model with salt2.4 files.

        Parameters
        ----------
        filename : str
            The original SALT2-4 file.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        f = np.load(filename)
        phase_grid = f['phase_grid']
        wl_grid = f['wl_grid']
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)
        
        xx = np.linspace(wl_grid[0], wl_grid[-1], basis.bx.nj)                               
        yy = np.linspace(phase_grid[0], phase_grid[-1], basis.by.nj)                          
        
        x, y = np.meshgrid(xx, yy)
        x, y = x.ravel(), y.ravel()
        jacobian = model.basis.eval(x, y)
        factor = cholesky_AAt(jacobian.T, beta=1e-20)

        return factor(jacobian.T * f['M0']), factor(jacobian.T * f['M1']), f['CL_pars']

    def pars_release(self):
        """
        Fixed the parameters label a pars_fix.
        """
        self.pars_free = [i for i in self.all_pars if i not in self.pars_fix]
        self.model.pars.fix()
        for pr in self.pars_free:
            self.model.pars.release(pr)

    def parameter_degradation(self, model_coef_init=0.05,
                              var_coef_init=0.1,
                              seed=None):
        """
        Degrade the parameters.
        Multiply their values by normal distribution, of mean 1 and a chosen standard deviations.
        For :math:`t_{max}` add normal distribution, of mean 0 and a chosen standard deviations.

        Parameters
        ----------
        model_coef_init : float
            Standard deviation for the model parameters : :math:`(X_0, X_1, c, t_{max})`
        var_coef_init : float
            Standard deviation for the error snake parameters : :math:`\gamma`
        seed : int
            Seed for numpy.random
        """
        if seed is not None:
            np.random.seed(seed)
            
        self.variance_model.pars.full[:] *= np.random.normal(1., var_coef_init,
                                                             len(self.variance_model.pars.full))
        print(f'\nVarModPars times N(1, {var_coef_init})')
        nb_sn = len(self.model.training_dataset.sne)
        for ps in self.pars_free:
            if (ps != 'eta_calib') & (ps != 'tmax'):
                mv = model_coef_init
                self.model.pars[ps].full[:] *= np.random.normal(1., mv,
                                                                len(self.model.pars[ps].full))
                if ps == 'X1':
                    self.model.pars[ps].full[:] -= self.model.pars[ps].full[:].mean()
                    
                print(f'{ps} times N(1, {mv})')

            elif model_coef_init != 0.:
                if ps == 'tmax':
                    s_tmax = 0.05
                    self.model.pars['tmax'].full[:] += np.random.normal(0, s_tmax, nb_sn)
                    print(f'{ps} + N(0, {s_tmax})')
        print('\n')
        self.gamma0 = self.variance_model.pars.free.copy()
        self.beta0 = self.model.pars.free.copy()

    def regularization(self, sur_reg=np.array(['M0', 'M1'])):
        """
        Instantiate regularization class.

        Parameters
        ----------
        sur_reg : numpy.array
            Surfaces to regularized.
        """
        sur_id = np.in1d(sur_reg, self.pars_free)
        if sur_id.sum() != 0:
            print(f'reg : {sur_reg[sur_id]}')
            self.reg = SplineRegul2D(self.model, mu=self.mu_reg,
                                     surfaces_reg=sur_reg[sur_id])
        else:
            self.reg = None

    def constraints(self, paramaters_cons=np.array(['M0', 'M1', 'CL', 'X0', 'tmax', 'X1', 'c'])):
        """
        Instantiate regularization class.

        Parameters
        ----------
        paramaters_cons : numpy.array
            List of parameters on which a constraint is applied.
        """
        cons_id = np.in1d(paramaters_cons, self.pars_free)

        # if model is fixed no need of contains
        if np.in1d(np.array(['M0', 'M1', 'CL']), self.pars_fix).sum() != 0.:
            self.cons = None

        elif cons_id.sum() != 0:
            cc = [i for i in paramaters_cons[cons_id]]
            print(f'cons : {cc}')
            self.cons = Constraints2D(self.model, mu_pq=self.mu_pq, parameters_cons=cc,
                                      mu_pq_var_x1=self.mu_pq_var_x1,
                                      x1_var_real=self.x1_var_real)   # ,vals = [])

        else:
            self.cons = None

    def __call__(self, pars_fix=[],
                 g=None, eta_covmatrix=None, sigma_kappa=None,
                 fit_spec=True, sigma_kappa_fit=True):
        """
        Fit !

        Parameters
        ----------
        pars_fix : list
            List of fixed parameters during the fit. If the error snake should
            be fixed 'gamma' should appear here.
        g : np.array
            Error snake parameters free or fixed.
        eta_covmatrix : float or numpy.array
            Calibration uncertainty matrix.
        sigma_kappa :
            Color scatter parameter.
        fit_spec : bool
            If spectral data are considered.
        sigma_kappa_fit : bool
            If color scatter should be fixed.

        TODO : add color scatter parameters fixation in pars_fix.
        """
        
        self.pars_fix = np.array(pars_fix)
        self.pars_release()

        print("\nNumber of free parameters :")
        for i in self.all_pars:
            print(i, self.model.pars[i].free.shape)
        print("\n")
        
        self.regularization()
        self.constraints()
        self.g = g
        if fit_spec is False:
            self.model.queue = [i for i in self.model.queue if (i.data['spec_id'].mean() == -1)]
            
        self.f = ModelPulls(self.model,
                            variance_model=self.variance_model, fit_spec=fit_spec,
                            chi2_normalization=self.chi2_normalization)
        
        if 'gamma' in pars_fix:
            self.f.VarianceModel.pars.fix()

        if sigma_kappa is not None:
            self.color_scatter = self.color_scatter_func(self.model.lambda_c_red, sigma_kappa) 
        else :
            self.color_scatter = None
        
        minimizer = self.minimize_func(self.f, constrains=self.cons,
                                       regularization=self.reg,
                                       color_scatter=self.color_scatter, ##
                                       n_iter=self.N_ITER, 
                                       df_stop=self.df_stop, log=self.log, beta_cholmod=self.BETA)
        
        minimizer(self.model.pars.free, sigma_kappa=sigma_kappa,
                  gamma0=self.g, eta_covmatrix=eta_covmatrix, sigma_kappa_fit=sigma_kappa_fit)

        self.M = minimizer.M
        self.gradf = minimizer.gradf
        self.minim = minimizer
        
        self.sigma_kappa = minimizer.sigma_kappa
        if (self.sigma_kappa is not None) & sigma_kappa_fit:
            self.dsigma_kappa = minimizer.dsigma_kappa
        
        self.timing = {'iter': np.array(minimizer.titer).mean(),
                       'cholesky': np.array(minimizer.tcholesky).mean(),
                       'model': np.array(minimizer.tmodel).mean(),
                       'hessian': np.array(minimizer.thessian).mean(),
                       's_iter': np.array(minimizer.titer).std(),
                       's_cholesky': np.array(minimizer.tcholesky).std(),
                       's_model': np.array(minimizer.tmodel).std(),
                       's_hessian': np.array(minimizer.thessian).std()}
        
        self.log = minimizer.log
        
        self.beta = minimizer.beta
        self.gamma = minimizer.gamma

        if fit_spec is False:
            self.model.queue = self.queue_wsp

    def control_plot(self, n_fit, variance_model=None,
                     theta_init=None, save_dir=None, add_name=''):
        """
        Make control plots.
        """
        if save_dir is not None:
            save_dir = f'{save_dir}'
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            
        if variance_model is not None:
            variance_model = self.variance_model
        if self.N_sn_4plot is None:
            self.N_sn_4plot = np.random.randint(0, len(self.training_dataset.sne), 5)
            
        for isn in self.N_sn_4plot:
            plots.plot_lc(self.model, isn, variance_model=variance_model,
                          save_dir=save_dir, add_name=add_name)
            plots.plot_spec(self.model, isn, variance_model=variance_model,
                            save_dir=save_dir, add_name=add_name)
        # plots.binning_residuals(self.model, save_dir=save_dir, add_name=add_name)
        plots.residual_histogram(self.model, n_fit, variance_model=variance_model,
                                 data_selection=False, save_dir=save_dir)

        if variance_model:
            plots.plot_2Dsurfacemodel(variance_model, ['gamma'],
                                      vmin=0.90, vmax=1.1, ddd=False,
                                      init_unit=True, theta_init=theta_init,
                                      save_dir=save_dir, add_name=add_name)






import numpy as np
from ..util.io import NTuple



class DataProxy:
    """ Provide flexible ntuple manipulation.

    In particular allows to map fields of a ntuple to arbitrary field
    names and eases the creation of index for fields.
    """
    def __init__(self,filename, **keys):
        """Parameters:
	-----------

	keys: provides a mapping between fields of the ntuple and
	arbitrary names. For exemple, keys mag='mmag' will create
	the view: dp.mag = nt['mmag']
	"""
	# Load data
        if isinstance(filename, str):
            self.nt = NTuple.fromfile(filename)
        else:
            self.nt = filename
        self.mapping = keys
        self._create_views()
        self._index_list=set()
        self._proxy_list=[]
        self._external_field_list=[]

    def _create_views(self):
        # Create views
        for k, v in self.mapping.items():
            setattr(self,k,self.nt[v])

    def add_views(self, **keys):
        """ A view is a mapping of a NTuple field: dp.key = nt['otherkeyname']"""
        for k, v in keys.items():
            setattr(self, k, self.nt[v])
        self.mapping.update(keys)

    def _update_proxys(self):
        for k, v in self._proxy_list:
            setattr(self,k,v(self))

    def add_proxy(self, name, f):
        self._proxy_list.append((name, f))
        setattr(self,name,f(self))

    def add_partial(self, name, data, index):
        #self._partial_list.append((name,data,index))
        s = np.zeros(index.max()+1)
        s[index.astype('int')] = data
        setattr(self, name, s)

    def add_field(self, name, data):
        """ Add external data sharing the same number of records.

	There is no mem copy so this is more memory efficient than
	just catenating the field to the existing NTuple.""" 
        self._external_field_list.append(name)
        setattr(self, name, data)

    def _update_external_fields(self, condition):
        for name in self._external_field_list:
            setattr(self, name, getattr(self,name)[condition])

    def compress(self, condition):
        self.nt = self.nt[condition].squeeze()
        self._create_views()
        self._update_external_fields(condition)
        self._update_proxys()
        self._update_index(condition)

    def make_index(self, fieldname, intmap=False):
        """ Build an index of the values taken by field fieldname.
        
        Exemples:
        ---------
        Let assume that self.mjd = [12,14,14]
        calling self.make_index('mjd') will create:
        self.mjd_set = [12,14]
        self.mjd_map = {12:1, 14:2}
        self.mjd_index = [1,2,2]
        """
        self.update_index(fieldname, intmap)
        self._index_list.add((fieldname, intmap))

    def update_index(self, fieldname, intmap=False):
        if intmap:
            a = getattr(self,fieldname).astype('int')
            s = set(a)
            s = list(s)
            s.sort()
            miv = a.min()
            mav = a.max()
            m = np.ones(mav-miv+1, dtype='int') * -1
            s = np.fromiter(s,dtype='int')
            for i, e in enumerate(s):
                m[e-miv] = i
            setattr(self, fieldname+'_set',s)
            setattr(self, fieldname+'_map',m)
            setattr(self, fieldname+'_index', m[a - miv])
        else:
            a = getattr(self,fieldname)
            s = set(a)
            s = np.array(list(s)) #np.fromiter(s, dtype=type(a[0]))
            m = dict(list(zip(s, list(range(len(s))))))
            setattr(self, fieldname+'_set',s)
            setattr(self, fieldname+'_map', m)
            i = np.fromiter((m[e] for e in a),'int')
            setattr(self, fieldname+'_index', i)

    def _update_index(self,condition):
        for fieldname, intmap in self._index_list:
            self.update_index(fieldname, intmap)
	#for ind in (getattr(self, i) for i in self._index_list):
	#    ind = ind[condition]



import numpy as np
import os
from .reindex import reindex
from ..dataset import TrainingDataset
from ..util import salt2


class Generator(object):
    r"""
    Data generator.
    Generate JLA simulation (Betoule et al. 2014)

    We start again from the objects composing each of these training batches~: we take their redshifts,
    their SALT2 parameters (:math:`X_0`, :math:`X_1`, :math:`c`, :math:`t_{max}`) as well as the
    photometric and spectroscopic observation dates.

    Attributes
    ----------
    filespath : str
        Path to the repository containing the light curve et spectra files.
    Sne ; numpy.array
        SNe names
    lc : numpy.rec.array
        Light curves data with data_type as type.
    sp : numpy.rec.array
        Spectral data, with data_type as type.
    snInfo : numpy.rec.array
        SNe information (:math:`z`, :math:`tmax`, :math:`x1`, :math:`x0`, :math:`c`)
    data_type : list
        NaCl data type.
    snInfo_type : list
        Type of information of SNe.
    sigma_lc : numpy.array
        Dispersion of photometric data
    sigma_sp : numpy.array
        Dispersion of spectral data.
    trainingDataset : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    """
    def __init__(self, filespath='../data/JLA_training_sample'):
        """
        Constructor - computes the arguments
        Args:
            filespath : string
               path to the repository containing the light curve et spectra files.
        """
        self.filespath = filespath

        self.Sne, self.snInfo, self.lc, self.sp = None, None, None, None
        self.sigma_sp, self.sigma_lc, self.trainingDataset, self.id_lc = None, None, None, None

        self.data_type = [('Date', '<f8'), ('Flux', '<f8'), ('FluxErr', '<f8'),
                          ('Filter', '|S20'), ('Wavelength', '<f8'), 
                          ('MagSys', '|S20'), ('ZHelio', '<f8'), ('sn_id', '<i4'), ('spec_id', '<i4')]
        self.snInfo_type = [('z', '<f8'), ('tmax', '<f8'), ('x1', '<f8'),
                            ('x0', '<f8'), ('c', '<f8')]
        self.make_data()

    def make_data(self):
        r"""
        Make training set from jla*.npy files, where all JLA data are gathered.
        Calculated ratio of error measured on measurements.
        
        Remove data out from definition ranges.
        """
        lc = np.load(self.filespath + os.sep + 'jla_lc.npy')
        lc_sum = np.load(self.filespath + os.sep + 'jla_lc_summary.npy', allow_pickle=True)
        sp = np.load(self.filespath + os.sep + 'jla_spectra.npy')
        sp_sum = np.load(self.filespath + os.sep + 'jla_spectra_summary.npy', allow_pickle=True)

        lc_sum = lc_sum[np.unique(lc['id'])]
        sp_sum = sp_sum[np.unique(sp['id'])]    
        self.Sne = lc_sum['SN'][np.unique(lc['id'])]
        lc['id'] = reindex(lc['id'])
        sp['id'] = reindex(sp['id'])

        lc_mean_wavelength = np.zeros_like(lc['id'])
        bands = [i.decode('UTF-8') for i in np.unique(lc['Filter'])]
        filterset = salt2.load_filters(bands)
        for bd in bands:
            idxmw = lc['Filter'] == bd.encode('UTF-8')
            lc_mean_wavelength[idxmw] = filterset.mean_wavelength([bd])

        # self.sne = lc_sum['SN']
        self.lc = np.rec.fromarrays((lc['Date'], lc['Flux'], lc['FluxErr'],
                                     lc['Filter'], lc_mean_wavelength,
                                     lc['MagSys'], lc_sum['ZHelio'][lc['id']], lc['id'],
                                     # len(np.unique(sp['id'])) + np.arange(len(lc['id']))
                                     -1 * np.ones_like(lc['Flux'])),
                                    dtype=self.data_type)
        sp_none = np.array([b''] * len(sp))
        sp_id = np.array([np.where(lc_sum['SN'] == i)[0][0] for i in sp_sum['SN']])[sp['id']]
        
        self.sp = np.rec.fromarrays((sp_sum['Date'][sp['id']], sp['Flux'],
                                     sp['FluxErr'], sp_none, sp['Wavelength'],
                                     sp_none, lc_sum['ZHelio'][sp_id], sp_id, sp['id']),
                                    dtype=self.data_type)

        self.sigma_lc = np.abs(self.lc['FluxErr'])/np.abs(self.lc['Flux'])
        self.sigma_sp = np.abs(self.sp['FluxErr'])/np.abs(self.sp['Flux'])
        
        self.snInfo = np.rec.fromarrays((lc_sum['ZHelio'], lc_sum['DayMax'], lc_sum['X1'],
                                        lc_sum['X0'], lc_sum['Color']), dtype=self.snInfo_type)
        self.indexing()

        # 392 spectre sn z = 0.68, obs date = 53404. MJD
        # 392 spectre sn z = 0.64, obs date = 53445. MJD
        print(" removing spectra 392,398 index by this confing")            
        # print('LEN SP : ', len(self.sp))
        self.sp, self.sigma_sp = self.remove()  # [392,398])
        # print('LEN SP : ', len(self.sp))
        # print('LEN SigP : ', len(self.sigma_sp))

        self.lc, self.sigma_lc, self.sp, self.sigma_sp = self.remove_sn_low_c()
        
        self.trainingDataset = TrainingDataset(self.lc, self.sp, self.snInfo)

    def remove(self):
        """
        Removing 2 spectra at late date.

        Returns
        -------
        sp : numpy.rec.array
            Spectra array re-indexed without 2 spectra
        sp_err :
            Flux error array re-indexed without 2 spectra
        """
        sp = self.sp.copy()
        sp_err = self.sigma_sp.copy()
        spec_id = []
        for iz, idd in zip([0.68, 0.64], [53404., 53445.]):
            spec_id.append(np.unique(sp['spec_id'][(sp['ZHelio'] == iz) & (sp['Date'] == idd)])[0])
        print('removed id spec_id : ', spec_id)
        spec_id = np.array(spec_id)
        for si in spec_id:
            # print(sp[sp['spec_id'] == si]['Date'][0], sp[sp['spec_id'] == si]['ZHelio'][0])
            idx = sp['spec_id'] == si
            sp = np.delete(sp, idx)
            sp_err = np.delete(sp_err, idx)
            sp['spec_id'][sp['spec_id'] > si] -= 1
            sp['i'][sp['spec_id'] > si] -= idx.sum()
            spec_id -= 1
        return sp, sp_err

    def remove_sn_low_c(self):
        """
        Removing SN with :math:`c < -1`

        Returns
        -------
        lc : numpy.rec.array
            Light curves array re-indexed those SNe.
        lc_err :
             Light curves array re-indexed without those SNe.
        sp : numpy.rec.array
            Spectra array re-indexed without those SNe.
        sp_err :
            Flux error array re-indexed without those SNe.
        """
        sp = self.sp.copy()
        sp_err = self.sigma_sp.copy()
        spec_id = []

        lc = self.lc.copy()
        lc_err = self.sigma_lc.copy()
        lc_id = []

        c = self.snInfo['c'] 
        idxxc = np.where(c < -1)[0]
        for idd in idxxc:
            spid = np.unique(sp['spec_id'][sp['sn_id'] == idd])
            spec_id.append(spid)
            lcid = np.unique(lc['lc_id'][lc['sn_id'] == idd])
            lc_id.append(lcid)

            idx_lc = lc['sn_id'] == idd
            lc = np.delete(lc, idx_lc)
            lc_err = np.delete(lc_err, idx_lc)
            
            lc['lc_id'][lc['sn_id'] > idd] -= len(lcid)
            lc['i'][lc['sn_id'] > idd] -= idx_lc.sum()
            lc['sn_id'][lc['sn_id'] > idd] -= 1

            idx_sp = sp['sn_id'] == idd
            sp = np.delete(sp, idx_sp)
            sp_err = np.delete(sp_err, idx_sp)
            
            sp['spec_id'][sp['sn_id'] > idd] -= len(spid)
            sp['i'][sp['sn_id'] > idd] -= idx_sp.sum()
            sp['sn_id'][sp['sn_id'] > idd] -= 1

        self.snInfo = np.delete(self.snInfo, idxxc)
        self.Sne = np.delete(self.Sne, idxxc)
        return lc, lc_err, sp, sp_err

    def indexing(self):
        """
        Reindex spectral ('spec_id', 'sn_id') and photometric data ('band_id', 'lc_id', 'sn_id'),
        when data have been removed.
        """
        n_lc, n_sp = len(self.lc), len(self.sp)
        # band indexation 
        dict_bd = {}
        _, idx_bd = np.unique(self.lc['Filter'], return_index=True)
        for i_bd, bd in enumerate(self.lc['Filter'][np.sort(idx_bd)]):
            dict_bd[bd] = i_bd
        id_bd = np.array([dict_bd[bd] for bd in self.lc['Filter']])

        # light curve indexation
        c = 0
        id_lc = np.ones(len(self.lc['Flux']))  # .astype(int)
        
        for i in range(self.lc['sn_id'][-1]+1):
            idx_sn = self.lc['sn_id'] == i
            lcs = self.lc[idx_sn]
            _, idx = np.unique(lcs["Filter"], return_index=True)
            for bd_sn in lcs['Filter'][np.sort(idx)]:
                id_lc[(self.lc['sn_id'] == i) & (self.lc['Filter'] == bd_sn)] = c  # [c]*len(lc[lc]))
                c += 1

        self.id_lc = np.hstack(np.array(id_lc))
        self.lc = np.lib.recfunctions.rec_append_fields(self.lc,
                                                        names=['lc_id', 'band_id', 'i'],
                                                        data=[id_lc.astype(int), id_bd,
                                                              np.arange(n_lc)])
        i_sp = n_lc + np.arange(n_sp)
        sp_ones = np.ones_like(i_sp)
        
        self.sp = np.lib.recfunctions.rec_append_fields(self.sp,
                                                        names=['lc_id', 'band_id', 'i'],
                                                        data=[-1*sp_ones, -1*sp_ones,
                                                              n_lc + np.arange(n_sp)])



import numpy as np
import scipy.optimize
import scipy.sparse
import scipy.sparse.linalg


def sqp(model, y, h, x0, n_iter=30, sigma=None):
    """ Solves a non-linear programming problem with equality constraints using a Gauss-Newton approximation of the Hessian

    The problem must write as follows:
    Find x that minimize: \sum (y_i - model(x)_i)^2
    subject to the constraint h(x) = 0
    
    Parameters:
    -----------
    model: callable with the signature (params, jac=False)
           must return if jac: the model evaluated at point params
                         else: the model and its jacobian under the form of a sparse matrix
           can raise ValueError in which case the model is assumed undefined at x
    y: array
       The fit data 
    h: callable with the same signature as model
    x0: array
        starting point
    n_iter: maximal number of iterations

    Returns:
    --------
    The best fit after n_iter iterations
    
    """
    n = len(x0)
    x = x0
    iter = 0
    alpha = 0.7
    
    while iter < n_iter:
        print(iter)

        val, J = model(x, jac=True)
        residuals = y - val
        if sigma is not None:
            J.data /= sigma[J.row]
            residuals /= sigma
        B = np.dot(J.T, J)
        
        gradf = -J.T * residuals
        
        if h is not None:
            cons, H = h(x, jac=True)
            M = scipy.sparse.bmat([[B, H.T], [H, None]])
            d0 = scipy.sparse.linalg.spsolve(M, -np.r_[gradf, cons])
            mu0 = d0[n:]
            d0 = d0[:n]
            obj_start = (residuals ** 2).sum() + (cons * mu0).sum()
            def phi(x):
                try:
                    if sigma is not None:
                        p = (((y - model(x)) / sigma) ** 2).sum() + (h(x) * mu0).sum()
                    else:
                        p = ((y - model(x)) ** 2).sum() + (h(x) * mu0).sum()
                except ValueError:
                    p = 1e12
                return p
        else:
            d0 = scipy.sparse.linalg.spsolve(B, -gradf)
            obj_start = (residuals ** 2).sum()
            def phi(x):
                if sigma is not None:
                    return (((y - model(x)) / sigma) ** 2).sum()
                else:
                    return ((y - model(x)) ** 2).sum()
        def crit(t):
            return phi(x + t * d0)

        t, fval, ni, funcalls = scipy.optimize.brent(crit, brack=(0, 1), full_output=True)

        dof = len(y) - len(x)
        print(('stepsize: %g, objective: %g -> %g, decrement: %g, D.o.F.: %d, objective/dof: %g' % (t, obj_start, fval, obj_start - fval, dof, fval / dof))) 
        
        x = x + t * d0
        
        iter = iter + 1
    return x





def check_cond(model, y, h, x0, sigma=None, worst=4, best=0):
    import matplotlib.pyplot as plt
    n = len(x0)
    x = x0
    iter = 0
    alpha = 0.7
    
    val, J = model(x, jac=True)
    residuals = y - val
    if sigma is not None:
        J.data /= sigma[J.row]
        residuals /= sigma
    B = np.dot(J.T, J)
        
    gradf = -J.T * residuals

    if h is not None:
        cons, H = h(x, jac=True)
        M = scipy.sparse.bmat([[B, H.T], [H, None]])
    else:
        M = B

    s, u = np.linalg.eigh(M.todense())

    print(("condition number: %g" % (abs(s).max() / abs(s).min())))
    plt.figure()
    plt.semilogy(abs(s))
    plt.ylabel('eigen values')

    plt.figure()
    reorder = np.argsort(abs(s))
    for i in range(worst):
        plt.plot(u[:,reorder[i]], label='%g' % s[reorder[i]])

    for i in range(best):
        plt.plot(u[:,-i], label='%g' % s[-i])

    plt.legend(loc='best')


















"""Access point to the instrument models.

At the moment, the module is still connected to
`saunerie.instruments`.  We will change that and probably stick to the
I/O system implemented in sncosmo.

"""

import numpy as np
from sksparse.cholmod import cholesky_AAt

import logging
import os
# TODO: import saunerie.instruments here
# TODO: maybe make it compatible with sncosmo I/O's
from .lib.instruments import InstrumentModel

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)


ZP_ERROR = {'MEGACAMPSF::g': 3,
            'MEGACAMPSF::i': 4,
            'MEGACAMPSF::r': 6,
            'MEGACAMPSF::z': 8,
            'SDSS::g': 4,
            'SDSS::i': 3,
            'SDSS::r': 2,
            'SDSS::u': 8,
            'SDSS::z': 5,
            'SWOPE::B': 8,
            'SWOPE::V': 8,
            'SWOPE::r': 8,
            'SWOPE::u': 23,
            '4SHOOTER2::B': 11,
            '4SHOOTER2::I': 20,
            '4SHOOTER2::R': 8,
            '4SHOOTER2::Us': 70,
            '4SHOOTER2::V': 7,
            'KEPLERCAM::B': 11,
            'KEPLERCAM::Us': 31,
            'KEPLERCAM::V': 7,
            'KEPLERCAM::i': 8,
            'KEPLERCAM::r': 25,
            'STANDARD::B': 15,
            'STANDARD::I': 15,
            'STANDARD::R': 15,
            'STANDARD::U': 100,
            'STANDARD::V': 15,
            'SWOPE2::B': 8,
            'SWOPE2::V': 8,
            'SWOPE2::g': 9,
            'SWOPE2::i': 8,
            'SWOPE2::r': 8,
            'SWOPE2::u': 23}

_instrument_repository = {}  # 'data/SALT3_training_sample/filters'}


def load_instrument(name, path=None):
    """
    Search for the requested instrument in a global dictionary.  If
    not found, load the instrument and return it (and update the
    cache).

    Parameters
        -------
        name : str
            Name of the instrument.
        path : None or str
            Whether instrument file should be specified.

        Returns
        -------
        ret : nacl.lib.instruments.InstrumentModel
            Instrument model.
    """
    global _instrument_repository
    ret = _instrument_repository.get(name, None)
    if ret is None:
        if path is None:
            ret = InstrumentModel(name)
        else:
            ret = InstrumentModel(name, path=path + os.sep + name)
        _instrument_repository[name] = ret
    return ret


_transmission_repository = {}


def load_transmission(instrument_name, passband_name, path=None):
    """
    Search for the specifier instrument and then the specified filter
    in their respective caches.  If not found, load, update the caches
    and return the passband.

    Parameters
    -------
    instrument_name : str
        Name of the instrument.
    passband_name : str
        Name of the desired band.
    path : None or str
        Whether instrument file should be specified.

    Returns
    -------
    ret : nacl.lib.instruments.InstrumentModel
        Filter transmission.

    """
    global _transmission_repository
    full_name = instrument_name + '::' + passband_name
    ret = _transmission_repository.get(full_name)
    if ret is None:
        try:
            instr = load_instrument(instrument_name, path=path)
            ret = instr.EffectiveFilterByBand(passband_name)
            #print('FLux in e not ADU')
            #ret = instr.get_efficiency_at_airmass(passband_name)
            _transmission_repository[full_name] = ret
        except KeyError:
            raise KeyError
    return ret


def load(full_name, path=None):
    """
    Transmission loader.

    Parameters
    -------
    full_name : str
        Names of the instrument & band such as 'INSTRUMENT::BAND'
    path : None or str
        Whether instrument file should be specified.

    Returns
    -------
    numpy.array
        Filter transmission.
    """
    instrument_name, passband_name = full_name.split('::')
    return load_transmission(instrument_name, passband_name, path=path)


class FilterDb:
    """
    Utility class that projects a set of filters on B-splines 
    and stores the projection coefficients. 
    
    The class is given a default projection basis, but it is always
    possible to define one (adapted) basis per filter.

    Attributes
    ----------
    basis : nacl.lib.bspline.BSpline
        Spline basis.
    _grid : np.array
        Grid, spline evaluation
    _J : scipy.sparse.csr_matrix
        Jacobian matrix.
    _factor : sksparse.cholmod.Factor
        Result of cholesky decomposition.
    filterpath : None or str
        Whether filter files should be specified.
    transmission_db : dict
        Transmission filters data base
    db : dict
        Data base of transmission project on basis is corrected from redshift.
    """

    def __init__(self, basis, tds=None, additional_band=None, filterpath=None):
        """
        Constructor - initialize the class, build a default projector.

        Parameters
        ----------
        basis : nacl.lib.bspline.BSpline
            Default BSpline basis on which to decompose the filters
        tds : nacl.dataset.TrainingDataset
            if present, load all the transmissions which
            are used in the training dataset and insert them in the database. [optional]
        """
        self.basis = basis
        self._grid, self._J, self._factor = self._projector(self.basis)
        self.filterpath = filterpath
        # I leave a trace of the factor that the FilterSet did actually
        # include the possibility of adding an external atmosphere model
        # Never used in practice -> I got rid of that.
        # self._atm = np.ones_like(self._grid)
        
        self.transmission_db = {}
        self.db = {}

        # if a tds is given, load all the transmissions
        # which appear in it. 
        if tds is not None:
            for tr in tds.get_all_transmissions():
                self.insert(tr)
        if additional_band is not None:
            for bd in additional_band:
                self.insert(load(bd))
                
    def __len__(self):
        """
        Given number of filter stored in database.
        """
        return len(self.db)

    @staticmethod
    def _projector(basis):
        r"""
        Precompute the elements and factorization of the fit matrix 
            
        .. math:: (J^T J)^{-1} J^T
        
        this saves repeated calls to cholesky_AAt when processing
        other filters.

        Parameters
        ----------
        basis : nacl.lib.bspline.BSpline
            Spline basis.

        Returns
        -------
        gxx : np.array
            Grid, spline evaluation
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix.
        factor : sksparse.cholmod.Factor
            Result of cholesky decomposition.
        """
        # refine the basis grid
        gx = basis.grid
        gxx = np.hstack((gx, 0.5*(gx[1:]+gx[:-1])))
        gxx.sort()
        # precompute the projector
        # (will save repeated calls to cholesky_AAt)
        jacobian = basis.eval(gxx).tocsr()
        factor = cholesky_AAt(jacobian.T)
        return gxx, jacobian, factor

    def insert(self, tr, z=0., basis=None):
        """
        Project the transmission `tr` on the spline basis and insert it
        into the database.
        
        If basis is None (which should be almost always the case) we
        use the default projector. Otherwise, we recompute a projector
        and use it.

        Parameters
        ----------
        tr :
            Filter transmission.
        z : float
            SN redshift.
        basis : nacl.lib.bspline.BSpline
            Spline basis.

        Returns
        -------
        tq : numpy.array
            Transmission for grid of wavelength redshift corrected.
        b : nacl.lib.bspline.BSpline
            Spline basis.
        """
        if basis is None:  # then, use the default basis
            grid, jacobian, factor = self._grid, self._J, self._factor
            b = self.basis
        else:
            grid, jacobian, factor = self._projector(basis)
            b = basis
            
        y = tr(grid * (1.+z))  # * atm
        tq = self._factor(self._J.T * y)

        full_name = tr.InstrumentName + '::' + tr.Band
        self.transmission_db[full_name] = tr
        self.db[(full_name, z)] = (tq, b)
        if z == 0.:
            self.db[full_name] = (tq, b)
        return tq, b
        
    def __getitem__(self, key):
        """
        Retrieve the coefficients and basis for the requested filter. 
        """
        if key not in self.db:
            raise KeyError
        return self.db[key]
    
    def plot(self, key):
        pass
    
